//******************************************************************************
//* File       : gString.cpp                                                   *
//* Author     : Mahlon R. Smith                                               *
//*              Copyright (c) 2011-14 Mahlon R. Smith, The Software Samurai   *
//*                 GNU GPL copyright notice located in gString.hpp            *
//* Date       : 27-Aug-2014                                                   *
//* Version    : (see gStringVersion string below)                             *
//*                                                                            *
//* Description: This class is used to convert between UTF-8 character         *
//* encoding and wchar_t character encoding. Statistical information for       *
//* analysis and display of the data is generated.                             *
//*                                                                            *
//* This is not a comprehensive string class, but a highly-specialized,        *
//* single-purpose string class for text encoding conversions. The gString     *
//* class was developed to support the NcDialog class, NCurses class, NcWindow *
//* classes and their descendants, and is a compromise between the full        *
//* functionality of Glib::ustring and the need to keep the NCurses, NcWindow, *
//* and NcDialog classes independent from third-party libraries.               *
//*                                                                            *
//* The gString class is a stand-alone class, that is, it requires no third-   *
//* party source or libraries beyond the standard C and C++ libraries found in *
//* all GNU compiler releases. Note however that the flag '_XOPEN_SOURCE'      *
//* must be defined.                                                           *
//*                                                                            *
//* Dynamic memory allocation is not performed within the gString class, and   *
//* The capacity of gString (see gsMAXCHARS and gsMAXBYTES) was chosen to      *
//* accomodate the maximum length of a path/filename string,                   *
//* i.e. 'PATH_MAX' (4096 bytes) as defined in the POSIX standard and found    *
//* in 'limits.h'.                                                             *
//*                                                                            *
//* This class is relatively fast and is useful for text data analysis and     *
//* manipulation, but is rather inefficient for data storage (although perhaps *
//* no more inefficient than std::string() or std::wstring()).                 *
//*                                                                            *
//* Developed using GNU G++ (Gcc v: 4.4.2 through 4.8.0)                       *
//*  under Fedora Release 12, Kernel Linux 2.6.32.26-175.fc12.i686.PAE         *
//*         through                                                            *
//*        Fedora Release 16, Kernel Linux 3.6.11-4.fc16.i686.PAE              *
//*  GNOME terminal 2.28.2 through 3.2.1                                       *
//*                                                                            *
//******************************************************************************
//* Version History (most recent first):                                       *
//*                                                                            *
//* v: 0.0.22 03-Jul-2014                                                      *
//*   - Add 'compose' constructor. Similar to the 'compose' member method,     *
//*     but immediately formats the input data during instantiation.           *
//*                                                                            *
//* v: 0.0.21 20-Apr-2014                                                      *
//*   - Fix string termination bug in 'append()'.                              *
//*   - Add 'insert()' method group to insert new text at specified offset     *
//*     in existing text.                                                      *
//*   - Overload 'append()' method for appending a single, wchar_t character.  *
//*   - Add 'operator!=' method.                                               *
//*                                                                            *
//* v: 0.0.20 22-Oct-2013                                                      *
//*   - Bug fix to differentiate between 'int' size and 'long int' size.       *
//*     Although these are the same on Wintel hardware, other platforms may    *
//*     have different widths for 'int' and 'long int'.                        *
//*   - Convert fixed-arg version of 'compose' method to a pair of variadic    *
//*     methods: one for a 'printf'-style formatting string, and one for a     *
//*     'wprintf'-style formatting string. Note, however, that the ANSI-C/C++  *
//*     specification does not yet support wide-character formatting strings,  *
//*     so the compiler cannot check validity of the wchar_t formatting string.*
//*   - Add the formatInteger() methods to create a fixed-width display string *
//*     representing a 16-bit, 32-bit or 64-bit integer. This functionality is *
//*     the wide version of a FormatInteger() method created for another       *
//*     project.                                                               *
//*                                                                            *
//* v: 0.0.19 17-Jun-2013                                                      *
//*   - Add methods shiftChars() and shiftCols().                              *
//*   - Add method clear() to reset all internal members to empty string.      *
//*   - Add an extension to the swprintf formatting options for the 'compose'  *
//*     method: %b , %B  (binary integer formatting) See binaryFormat() method.*
//*   - Modify the 'compose' method to return a const pointer to the wchar_t   *
//*     data. This allows the result of the compose method to be directly      *
//*     inserted into the std::wcout output stream. (It's cool!)               *
//*                                                                            *
//* v: 0.0.18 17-Apr-2013                                                      *
//*   - Compensate for a logical error in the standard library swprintf()      *
//*     function string conversions. See notes in composeFieldwidthBugFix()    *
//*     method.                                                                *
//*                                                                            *
//* v: 0.0.17 11-Mar-2012                                                      *
//*   - Due to changes in the ncurses library between v:5.7 and v:5.9, it has  *
//*     become necessary to pass only one wide character at a time to          *
//*     wadd_wch() via the ncout_t class (cchar_t structure).                  *
//*     For this reason, it has become unnecessary and inefficient to          *
//*     pre-populate an array of ncout_t objects for output to the display.    *
//*     Therefore, all support for ncurses character-output 'chunks' has been  *
//*     removed. In addition, to achieve full independence from ncurses, we    *
//*     have removed the data member 'colorAttr' (color attribute) and the     *
//*     associated methods. These changes yield a savings of 2,600 bytes of    *
//*     code space in the gString class, and a savings of 6,100 bytes of data  *
//*     per gString instance.                                                  *
//*                                                                            *
//* v: 0.0.16 15-Jan-2012                                                      *
//*   - For methods that do not change the internal data members, make the     *
//*     methods 'const' to so indicate.                                        *
//*   - Fix bug in limitCharacters(). gsChars member was not counting the      *
//*     NULLCHAR in the character total.                                       *
//*                                                                            *
//* v: 0.0.15 02-Oct-2011                                                      *
//*   - Convert the conversion methods to use the reentrant, restartable       *
//*     mbsrtowcs() and wcsrtombs() library routines, rather than the older    *
//*     mbstowcs() and wcstombs().                                             *
//*   - Implement error-recovery methods Bytewise_u2w() and Bytewise_w2u().    *
//*     These methods are called if the string conversion fails due to an      *
//*     invalid multi-byte UTF-8 character or invalid wchar_t character,       *
//*     respectively. Converts the input up to the offending character,        *
//*     null-terminates both wchar_t and UTF-8 strings, and discards remaining *
//*     input.                                                                 *
//*                                                                            *
//* v: 0.0.14 19-Sep-2011                                                      *
//*   - Add 'append' methods for both UTF-8 and wchar_t data.                  *
//*                                                                            *
//* v: 0.0.13 16-Aug-2011                                                      *
//*   - Add a constructor and 'operator=' for creating a formatted string      *
//*     using gsForm class to pass the parameters.                             *
//*   - Add 'compose()' method for creating formatted strings. This uses the   *
//*     gsForm class implicitly.                                               *
//*                                                                            *
//* v: 0.0.12 01-Jul-2011                                                      *
//*   - Add array of character column width values. Can be used to facilitate  *
//*     scrolling through data.                                                *
//*   - Modified the Reinit() method so that a NULL string is properly         *
//*     represented as having gsChars==1 and utfBytes==1.                      *
//*   - Handle the special case of a constructor or operator= method receiving *
//*     a NULL string.                                                         *
//*   - Made gsout() method public even though it should only be called by the *
//*     NCurses and NcWindow class output routines. This was done to           *
//*     fully-protect the gString class data members from the clumsy NCurses   *
//*     class and NcWindow class developer, namely me. :-)                     *
//*   - Add stubs for data-shift and substring methods to facilitate dynamic   *
//*     edit of text strings in NcDialog-class controls. (not currently used)  *
//*                                                                            *
//* v: 0.0.11 16-Apr-2011                                                      *
//*   - Protect constructors and assignment operators from NULL-pointer        *
//*     parameters.                                                            *
//*                                                                            *
//* v: 0.0.10 20-Mar-2011                                                      *
//*   - Bump the version number to indicate that the basic functionality is    *
//*     complete and moderately-well tested.                                   *
//*                                                                            *
//* v: 0.0.01 06-Mar-2011                                                      *
//*   - First try at defining and implementing the needed functionality.       *
//*   - To maintain independence from curses.h, we have created our own        *
//*     'typedef ncattr unsigned int' data type for color attributes, rather   *
//*     than use the 'attr_t' typedef in the ncurses library.                  *
//*   - For the same reason we have defined the ncout_t class to resemble the  *
//*     ncurses 'struct cchar_t'. If some huge change in the ncurses library   *
//*     occurs (unlikely), we will have to revisit this design decision.       *
//******************************************************************************
//* Programmer's Notes:                                                        *
//* 1) Note that the isAscii data member is not initialized until the isASCII()*
//*    method is called; therefore, do not reference the isAscii data member   *
//*    within any class method without a previous call to isASCII().           *
//* 2) This class could be significantly expanded with a substr(), and other   *
//*    methods; however, such methods are not supported at this time.          *
//*    If a full-featured string class that understands UTF-8 encoded data is  *
//*    needed, we recommend the glibmm Glib::ustring class which is actually   *
//*    rather sexy...                                                          *
//*                                                                            *
//*                                                                            *
//******************************************************************************

//*****************
//* Include Files *
//*****************
#include "gString.hpp"     //* Class definition

//* ENABLE FOR DEBUGGING ONLY - NOT USED IN PRODUCTION CODE *
//#include <iostream>        //* Standard I/O definitions
//#include <fstream>         //* File I/O definitions
//using namespace std ;
//const char* gsDebugFile = "gStringDebug.txt" ;
// EXAMPLE DEBUG OUTPUT
//ofstream ofs ( gsDebugFile, ofstream::out | ofstream::app ) ;
//if ( ofs.is_open() )                         // if output file open
//{
//   ofs << "Conversion Error in ConvertUTF8toWide()" << endl ;
//   char b[128] ;
//   sprintf ( b, "ps.count:%02d  ps.value:%08X  %02hX %02hX %02hX %02hX",
//             ps.__count, ps.__value.__wch,
//             ps.__value.__wchb[0], ps.__value.__wchb[1], 
//             ps.__value.__wchb[2], ps.__value.__wchb[3] ) ;
//   ofs << b << endl ;
//   ofs << "Truncated string: \"" << this->uStr << "\"" << endl ;
//   ofs.close() ;
//}

//*********************
//* Local Definitions *
//*********************
#define ZERO (0)
#define NULLCHAR ('\0')
#define TILDE ('~')
#define SPACE (' ')

const short maxBPC = 6 ;      // maximum number of bytes in a UTF-8 character

//*********************
//* Local Data        *
//*********************

static const char gStringVersion[] = "0.0.22" ; //* Class version number

//* Enable/disable binary output extension to swprintf function *
#define BINARY_EXTENSION (1)

//* Enable/disable fixed-width-field integer formatting *
#define IFORMAT_EXTENSION (1)
#if IFORMAT_EXTENSION != 0
//* Length of an output buffer for infFormatXX() methods *
const short FI_BUFFER_SIZE = (16) ;
#endif   // IFORMAT_EXTENSION

#if ENABLE_GS_DEBUG != 0
static wchar_t dbmsg[gsMAXCHARS] ;  // for debugging only
#endif   // ENABLE_GS_DEBUG



//*************************
//*      ~gString         *
//*************************
//******************************************************************************
//* Destructor.                                                                *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************

gString::~gString ( void )
{

   // nothing to be done, currently

}  //* End ~gString() *

//*************************
//*       gString         *
//*************************
//******************************************************************************
//* Constructor: Convert specified UTF-8-encoded source to gString.            *
//*                                                                            *
//*                                                                            *
//* Input  : usrc  : pointer to a UTF-8-encoded, null-terminated string        *
//*          charLimit: (optional, gsMAXCHARS by default)                      *
//*                  maximum number of characters from source array to convert *
//*                     if a character limit is specified, the count           *
//*                     should not include the NULL terminator character.      *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

gString::gString ( const char* usrc, short charLimit )
{
   if ( charLimit <= ZERO )            // prevent stupidity
      charLimit = 1 ;
   else if ( charLimit > gsMAXCHARS )
      charLimit = gsMAXCHARS ;

   //* Reinitialize all data members and convert *
   //* source text from UTF-8 to wchar_t.        *
   this->Reinit() ;
   this->ConvertUTF8toWide ( usrc, charLimit ) ;

}  //* End gString() *

//*************************
//*       gString         *
//*************************
//******************************************************************************
//* Constructor: Convert specified wchar_t ('wide') source to gString.         *
//*                                                                            *
//*                                                                            *
//* Input  : wsrc  : pointer to a wchar_t-encoded, null-terminated string      *
//*          charLimit: (optional, gsMAXCHARS by default)                      *
//*                  maximum number of characters from source array to convert *
//*                     if a character limit is specified, the count           *
//*                     should not include the NULL terminator character.      *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* Programmer's Note: The following instantiation of gs2 works under the GNU  *
//* compiler but probably shouldn't:   gString gs1( "Hello!" ), gs2( gs1 ) ;   *
//******************************************************************************

gString::gString ( const wchar_t* wsrc, short charLimit )
{
   if ( charLimit <= ZERO )            // prevent stupidity
      charLimit = 1 ;
   else if ( charLimit > gsMAXCHARS )
      charLimit = gsMAXCHARS ;

   //* Reinitialize all data members and convert *
   //* source text from wchar_t to UTF-8.        *
   this->Reinit() ;
   if ( wsrc != NULL && *wsrc != NULLCHAR )
      this->ConvertWidetoUTF8 ( wsrc, charLimit ) ;

}  //* End gString() *

//*************************
//*       gString         *
//*************************
//******************************************************************************
//* Constructor: Convert formatting specification and its arguments to gString.*
//*              Please refer to compose() method for more information.        *
//*                                                                            *
//*                                                                            *
//* Input  : fmt  : a format specification string in the style of sprintf(),   *
//*                 swprintf() and related formatting C/C++ functions.         *
//*          arg1 : pointer to first value to be converted by 'fmt'            *
//*          ...  : optional arguments (between ZERO and gsfMAXARGS - 1)       *
//*                 Each optional argument is a POINTER (address of) the value *
//*                 to be formatted.                                           *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

gString::gString ( const char* fmt, const void* arg1, ... )
{
   //* Initialize the formatting structure.                       *
   //* Point to the formatting string and count the format specs. *
   gString gsfmt( fmt ) ;
   gsForm gsf { gsfmt.gstr() } ;
   short specCount = ZERO ;
   for ( short i = ZERO ; gsf.fmtSpec[i] != NULLCHAR ; i++ )
   {
      if ( gsf.fmtSpec[i] == L'%' && gsf.fmtSpec[i+1] != L'%' )
         ++specCount ;
      /* TEMP - SEE IMPLEMENTATION NOTES IN HEADER OF 'compose' METHOD */
      if ( specCount > gsFormMAXARGS ) { specCount = gsFormMAXARGS ; break ; }
   }

   //* Copy the source-data pointers to our formatting structure. *
   short i = ZERO ;              // target-argument index
   gsf.argList[i++] = arg1 ;     // get the fixed argument
   va_list argList ;
   va_start ( argList, ZERO ) ;
   for ( ; i < specCount ; i++ ) // get contents of variable-length arg list
      gsf.argList[i] = va_arg ( argList, void* ) ;
   va_end ( argList ) ;

   //* Convert the formatting instructions into formatted text data *
   wchar_t  wsrc[gsMAXCHARS] ;
   this->Convert_gsForm2gs ( gsf, wsrc ) ;

   //* Reinitialize all data members and convert *
   //* source text from wchar_t to UTF-8.        *
   this->Reinit() ;
   if ( *wsrc != NULLCHAR )
      this->ConvertWidetoUTF8 ( wsrc ) ;

}  //* End gString() *

//*************************
//*       gString         *
//*************************
//******************************************************************************
//* Constructor: Convert formatting instructions in gsForm class object        *
//*              to gString. See compose() for more information.               *
//*                                                                            *
//*                                                                            *
//* Input  : gsf   : initialized gsForm class object containing parameters for *
//*                  creating a formatted text string.                         *
//*          charLimit: (optional, gsMAXCHARS by default)                      *
//*                  maximum number of characters from source array to convert *
//*                     if a character limit is specified, the count           *
//*                     should not include the NULL terminator character.      *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* Output conversions: The specifier '%' is follwed by (optional) modifiers,  *
//* and then the data type for the conversion.                                 *
//* (see man pages: libc.info 12.12.3 - Table of Output Conversions)           *
//*                                                                            *
//******************************************************************************

gString::gString ( const gsForm& gsf, short charLimit )
{
   //* Convert the formatting instructions into formatted text data *
   wchar_t  wsrc[gsMAXCHARS] ;
   this->Convert_gsForm2gs ( gsf, wsrc ) ;

   //* Reinitialize all data members and convert *
   //* source text from wchar_t to UTF-8.        *
   this->Reinit() ;
   if ( *wsrc != NULLCHAR )
      this->ConvertWidetoUTF8 ( wsrc, charLimit ) ;

}  //* End gString() *

//*************************
//*       gString         *
//*************************
//******************************************************************************
//* Constructor: Initialize members to default values (NULL string).           *
//*                                                                            *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

gString::gString ( void )
{
   this->Reinit() ;                 // initialize members

}  //* End gString() *

//*************************
//*        clear          *
//*************************
//******************************************************************************
//* Reset contents to an empty string i.e. "". The data will consist of a      *
//* single, NULLCHAR character. The character and byte counts are set to       *
//* 1 (one), and the column count is zero.                                     *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void gString::clear ( void )
{

   this->Reinit () ;

}  //* End clear() *

//*************************
//*       Reinit          *
//*************************
//******************************************************************************
//* Reinitialize the data members to indicate a NULL string "\0".              *
//*                                                                            *
//*                                                                            *
//* Input  : initAll: (optional, 'true' by default)                            *
//*                   if 'true' reinitialize all data members                  *
//*                   if 'false' reinitialize everything EXCEPT:               *
//*                    gStr[] and gsChars.                                     *
//* Returns: nothing                                                           *
//******************************************************************************

void gString::Reinit ( bool initAll )
{

   if ( initAll != false )
   {
      this->gStr[ZERO] = NULLCHAR ;
      this->gsChars = 1 ;  // A NULL string has 1 character
   }
   this->uStr[ZERO] = NULLCHAR ;
   this->cWidth[ZERO] = ZERO ;
   this->utfBytes = 1 ;    // a NULL string has 1 byte
   this->gsCols = ZERO ;   // and occupies no columns
   this->isAscii = true ;  // a NULL string is an ASCII string

}  //* End Reinit() *

//*************************
//*     operator=         *
//*************************
//******************************************************************************
//* Assignment operator: converts UTF-8-encoded source to gString.             *
//*                                                                            *
//*                                                                            *
//* Input  : pointer to an array of UTF-8-encoded characters                   *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void gString::operator= ( const char* usrc )
{

   //* Reinitialize all data members and convert *
   //* source text from UTF-8 to wchar_t.        *
   this->Reinit() ;
   this->ConvertUTF8toWide ( usrc ) ;

}  //* End operator=() *

//*************************
//*     operator=         *
//*************************
//******************************************************************************
//* Assignment operator: converts wchar_t ('wide') source to gString.          *
//*                                                                            *
//*                                                                            *
//* Input  : pointer to an array of wchar_t 'wide' characters                  *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void gString::operator= ( const wchar_t* wsrc )
{

   //* Reinitialize all data members and convert *
   //* source text from wchar_t to UTF-8.        *
   this->Reinit() ;
   if ( wsrc != NULL && *wsrc != NULLCHAR )
      this->ConvertWidetoUTF8 ( wsrc ) ;

}  //* End operator=() *

//*************************
//*     operator=         *
//*************************
//******************************************************************************
//* Assignment operator: Converts gsForm-class instructions to gString.        *
//*                                                                            *
//*                                                                            *
//* Input  : an initialized gsForm object (by reference)                       *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void gString::operator= ( const gsForm& gsf )
{
   //* Convert the formatting instructions into formatted text data *
   wchar_t  wsrc[gsMAXCHARS] ;
   this->Convert_gsForm2gs ( gsf, wsrc ) ;

   //* Reinitialize all data members and convert *
   //* source text from wchar_t to UTF-8.        *
   this->Reinit() ;
   if ( *wsrc != NULLCHAR )
      this->ConvertWidetoUTF8 ( wsrc ) ;

}  //* End operator=() *

//*************************
//*     operator=         *
//*************************
//******************************************************************************
//* Assignment operator. Copies one gString object to another.                 *
//*                                                                            *
//*                                                                            *
//* Input  : gString object to be copied (by reference)                        *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void gString::operator= ( const gString& gssrc )
{

   this->Reinit() ;        // initialize members (not necessary, but comforting)

   this->gsCols    = gssrc.gsCols ;
   this->gsChars   = gssrc.gsChars ;
   this->utfBytes  = gssrc.utfBytes ;
   this->isAscii   = gssrc.isAscii ;
   short i = -1 ;
   do                               // wchar_t array and cWidth array
   {
      ++i ;
      this->gStr[i]   = gssrc.gStr[i] ;
      this->cWidth[i] = gssrc.cWidth[i] ;
   }
   while ( this->gStr[i] != NULLCHAR ) ;
   i = -1 ;
   do                               // char array
   {
      ++i ;
      this->uStr[i] = gssrc.uStr[i] ;
   }
   while ( this->uStr[i] != NULLCHAR ) ;

}  //* End operator=() *

//*************************
//*       compose         *
//*************************
//******************************************************************************
//* Create formatted text data from a format specification string including    *
//* between ZERO and gsfMAXARGS format specifications and their corresponding  *
//* argument pointers..                                                        *
//*                                                                            *
//* Supported data types:                                                      *
//*  %d, %i  integer (decimal)                                                 *
//*  %o      integer (octal)                                                   *
//*  %u      integer (unsigned)                                                *
//*  %x, %X  integer (hex lower or upper case)                                 *
//*  %f      floating point (fixed point)                                      *
//*  %e, %E  floating point (scientific notation, lower/uppercase)             *
//*  %g, %G  floating point (normal/exponential, lower/uppercase)              *
//*  %a, %A  floating point (hex fraction)                                     *
//*  %c      character                                                         *
//*  %C      character (alias for %lc)                                         *
//*  %s      string                                                            *
//*  %S      string (alias for %ls)                                            *
//*  %p      pointer                                                           *
//*  %b, %B  (extension to swprintf - see description below)                   *
//*  %m      capture 'errno' description string (see /usr/include/errno.h)     *
//*  %n      number of characters printed so far                               *
//*          (value written to corresponding argument's location)              *
//*  %%      literal '%'                                                       *
//*                                                                            *
//* See man pages for the C/C++ function 'swprintf' or                         *
//* 'Table of Output Conversions' for additional details.                      *
//* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  *
//*                                                                            *
//* Input  : fmt  : a format specification string in the style of sprintf(),   *
//*                 swprintf() and related formatting C/C++ functions.         *
//*          ...  : optional arguments (between ZERO and gsfMAXARGS)           *
//*                 Each optional argument is a POINTER (address of) the value *
//*                 to be formatted.                                           *
//*                 - Important Note: There must be AT LEAST as many optional  *
//*                   arguments as the number of format specifiers defined in  *
//*                   the formatting string. Excess arguments will be ignored; *
//*                   HOWEVER, too few arguments will result in an application *
//*                   crash and ridicule from your peers.                      *
//*                                                                            *
//* Returns: const wchar_t* to formatted data                                  *
//******************************************************************************
//* Implementation Notes:                                                      *
//* =====================                                                      *
//*                                                                            *
//* 1) Unsupported data types                                                  *
//*    ======================                                                  *
//*    Conversion modifiers that are not fully supported at this time:         *
//*                       'j', 'z', 't', '%['                                  *
//*    Note: The '*' field-width specification or precision specification      *
//*          which uses the following argument as the width/precision value    *
//*          IS NOT supported.                                                 *
//*                                                                            *
//* 2) Binary-format specifier                                                 *
//*    =======================                                                 *
//*    We implement an extension to the swprintf output-conversion-specifiers  *
//*    for binary formatted output. We have found this formatting option useful*
//*    when working with bit masks and for verifying bit-shifting operations.  *
//*    - Base formatting specifier:                                            *
//*       %b , %B                                                              *
//*    - Format modifiers for DATA SIZE are the same as for swprintf:          *
//*       hh , h , l , ll , L , q                                              *
//*    - Format modifier for prepending of a type indicator.                   *
//*       '#' (hash character)                                                 *
//*       This is the same principle as for prepending a '0x' indicator to hex *
//*       output, and will place either a 'b' or 'B' character at the beginning*
//*       of the output. Examples:  %#hhb -> b0101.1010   %#hhB -> B0101.1010  *
//*    - Format modifier for appending of a type indicator.                    *
//*       '-#' (minus sign and hash character)                                 *
//*       Rather than prepending the indicator, the indicator will be append   *
//*       to the end of the output.                                            *
//*       Examples: %-#hhb -> 0101.1010b   %-#hhB -> 0101.1010B                *
//*    - Format modifier for specifying the group-seperator character.         *
//*       By default, the bit groups are seperated by a '.' (fullstop)         *
//*       character. To specify an alternate seperator character:              *
//*       % hB -> 0111 0101 1010 0001    (' ' (space) character as seperator)  *
//*       %_hB -> 0111_0101_1010_0001    ('_' (underscore) char as seperator)  *
//*       %#/hB -> B0111/0101/1010/0001  ('/' (slash) character as seperator)  *
//*       %-#-hB -> 0111-0101-1010-0001B ('-' (dash) character as seperator)   *
//*       Valid seperator characters are any printable ASCII character         *
//*       that IS NOT alphabetical, IS NOT a number, and IS NOT a '.'(fullstop)*
//*    - Format modifier for specifying bit grouping.                          *
//*       By default, bits are formatted in groups of four (4 nybble); however,*
//*       if desired, bits can be formatted in groups of eight (8 byte):       *
//*       %.8hB -> 01110101.10100001                                           *
//*       %-.8hB -> 01110101-10100001                                          *
//*       %# .8hB -> B01110101 10100001                                        *
//*       %-#`.8hb -> 01110101`10100001b                                       *
//*                                                                            *
//* 3) Format-specification Deficiencies                                       *
//*    =================================                                       *
//*    The compiler will scan your UTF-8 formatting string for valid           *
//*    formatting specifications and warn you of potential problems.           *
//*      HOWEVER:                                                              *
//*    The ANSI-C/C++ specification does not yet support parsing of wchar_t    *
//*    (swprintf-style) formatting strings. If you pass a wchar_t formatting   *
//*    string, the compiler will not be able to warn of potential problems,    *
//*    so construct your formatting string carefully. See gcc documentation    *
//*    for more information:  info -f gcc.info -n 'Function Attributes'        *
//*      ADDITIONALLY:                                                         *
//*    Because the optional parameters are passed as POINTERS, similar to the  *
//*    C/C++ library function 'sscanf' and friends, the compiler cannot perform*
//*    automatic promotions from short int* to int* or from float* to double*, *
//*    and so-on. For this reason, please use care that the data type you      *
//*    specify in the formatting string matches the data type of the variable  *
//*    referenced by its parameter pointer. For example, the conversion        *
//*    specifications:  %d, %i, %u, %x, and %X  assume a parameter of 'int'    *
//*    size If your actual variable declaration is for a 8-bit int, short int, *
//*    long int or long long int, then explicitly indicate this in the         *
//*    corresponding conversion specification.                                 *
//*    Examples:                                                               *
//*      char      Greeting[] = { "Hello!" } ;                                 *
//*      int       iValue ;                                                    *
//*      long long int qValue ;                                                *
//*      long int  lValue ;                                                    *
//*      short int sValue1, sValue2, sValue3 ;                                 *
//*      bool      flagValue ;                                                 *
//*      float     fltValue ;                                                  *
//*      double    dblValue ;                                                  *
//*      gString gs ;                                                          *
//*      gs.compose( "%s - %d %12hd, %-hi, %#hx %08lXh %llu %hhd",             *
//*                  Greeting, &iValue, &sValue1, &sValue2, &sValue3,          *
//*                  &lValue, &qValue, &flagValue );                           *
//*      gs.compose( "floating downstream:%10.2f and doubling our pace:%.4lf", *
//*                  &fltValue, &dblValue ) ;                                  *
//*                                                                            *
//*      Note: Use care when formatting an 8-bit integer: 'char' ,             *
//*            'unsigned char' or a boolean value declared with the 'bool'     *
//*            keyword. For these types, use the 'hh' conversion modifier.     *
//*            Examples: %hhd , %hhu , %hhX                                    *
//*                                                                            *
//* 4) 'swprintf' bug-fix                                                      *
//*    ==================                                                      *
//*    The standard library 'swprintf' function has a design flaw for format   *
//*    specifications that include a field-width specifier.                    *
//*    'swprintf' pads the string to the specified number of CHARACTERS, not   *
//*    the number of COLUMNS as it should do. For (Arabic) numeric source      *
//*    values this is not a problem because one character == one display       *
//*    column. For string source data, however, if the source string contains  *
//*    characters that require more than one display column each, then the     *
//*    output may be too wide.                                                 *
//*    Therefore, for string-source-formatting specifications ONLY             *
//*          (examples: "%12s"  "%-6s"  "%16ls"  "%5S"  "%-24S")               *
//*    we compensate for this anomalous behavior by interpreting the           *
//*    field-width specifier as number-of-columns, NOT number-of-characters.   *
//*    This will result in output that appears different (and better) than     *
//*    output created directly by the 'swprintf' function.                     *
//*                                                                            *
//* 5) Parameter type checking                                                 *
//*    =======================                                                 *
//*    Unfortunately, type-checking of wchar_t formatting strings is not yet   *
//*    supported by the gnu (v:4.8.0) compiler, (but see wchar.h which is      *
//*    preparing for future expansion). Thus, use care when constructing your  *
//*    'wchar_t fmt' formatting string. The 'char fmt' string is type-checked. *
//*                                                                            *
//******************************************************************************
//* Programmer's Notes:                                                        *
//*  See gcc.info -n 'Function Attributes' for the __attribute__ keyword and   *
//*  the format (ARCHETYPE, STRING-INDEX, FIRST-TO-CHECK); definition.         *
//*  where:                                                                    *
//*    ARCHETYPE      is the built-in compiler template to use for typechecking*
//*    STRING-INDEX   is the (1-based) index of the formatting string          *
//*                   (there is an invisible zero-th argument, so formatting   *
//*                   (string index is '2')                                    *
//*    FIRST-TO-CHECK is the (1-based) index of the first argument to check    *
//*                   against the formatting string (a zero '0' for this       *
//*                   parmeter indicates that args should not be checked)      *
//*                                                                            *
//* NOTE: INITIALLY, WE USE THE EXISTING CONVERSION CODE.                      *
//*       LATER, WE CAN HAVE AN ARBITRARY NUMBER OF SPECS/ARGS.                *
//*       THIS REQUIRES RE-DEFINITION OF THE gsForm CLASS TO USE DYNAMIC       *
//*       ALLOCATION FOR void** args[]; AND A COUNT OF POINTERS IN THE LIST.   *
//*                                                                            *
//******************************************************************************

const wchar_t* gString::compose ( const wchar_t* fmt, ... )
{
   //* Initialize the formatting structure.                               *
   //* Point to the wchar_t formatting string and count the format specs. *
   gsForm gsf { fmt } ;
   short specCount = ZERO ;
   for ( short i = ZERO ; gsf.fmtSpec[i] != NULLCHAR ; i++ )
   {
      if ( gsf.fmtSpec[i] == L'%' && gsf.fmtSpec[i+1] != L'%' )
         ++specCount ;
      /* TEMP - SEE NOTE ABOVE */
      if ( specCount > gsFormMAXARGS ) { specCount = gsFormMAXARGS ; break ; }
   }

   //* Copy the source-data pointers to our formatting structure.         *
   va_list argList ;
   va_start ( argList, ZERO ) ;
   for ( short i = ZERO ; i < specCount ; i++ )
      gsf.argList[i] = va_arg ( argList, void* ) ;
   va_end ( argList ) ;

   //* Convert the formatting instructions into formatted text data *
   wchar_t  wsrc[gsMAXCHARS] ;
   this->Convert_gsForm2gs ( gsf, wsrc ) ;

   //* Reinitialize all data members and convert *
   //* source text from wchar_t to UTF-8.        *
   this->Reinit() ;
   if ( *wsrc != NULLCHAR )
      this->ConvertWidetoUTF8 ( wsrc ) ;

   return this->gStr ;

}  //* End compose() *

const wchar_t* gString::compose ( const char* fmt, ... )
{
   //* Formatting string is UTF-8 so that compiler can check the call     *
   //* arguments; however, we need it to be wchar_t.                      *
   gString gs( fmt ) ;

   //* Point to the wchar_t formatting string and count the format specs. *
   gsForm gsf { gs.gstr() } ;
   short specCount = ZERO ;
   for ( short i = ZERO ; gsf.fmtSpec[i] != NULLCHAR ; i++ )
   {
      if ( gsf.fmtSpec[i] == L'%' && gsf.fmtSpec[i+1] != L'%' )
         ++specCount ;
      /* TEMP - SEE NOTE ABOVE */
      if ( specCount > gsFormMAXARGS ) { specCount = gsFormMAXARGS ; break ; }
   }
   //* Copy the source-data pointers to our formatting structure.         *
   va_list argList ;
   va_start ( argList, ZERO ) ;
   for ( short i = ZERO ; i < specCount ; i++ )
      gsf.argList[i] = va_arg ( argList, void* ) ;
   va_end ( argList ) ;

   //* Convert the formatting instructions into formatted text data *
   wchar_t  wsrc[gsMAXCHARS] ;
   this->Convert_gsForm2gs ( gsf, wsrc ) ;

   //* Reinitialize all data members and convert *
   //* source text from wchar_t to UTF-8.        *
   this->Reinit() ;
   if ( *wsrc != NULLCHAR )
      this->ConvertWidetoUTF8 ( wsrc ) ;

   return this->gStr ;

}  //* End compose() *

//*************************
//*     formatInteger     *
//*************************
//******************************************************************************
//* Convert an integer value into a formatted display string of the specified  *
//* width. Value is right-justified in the field, with leading spaces added    *
//* if necessary (but see 'lJust' parameter).                                  *
//* Maximum field width is FI_MAX_FIELDWIDTH. This is wide enough to display   *
//* 10 decimal digits in comma-formatted form: '1,234,567,890'                 *
//*                                                                            *
//* Actual formatting of the value depends on the combination of value size    *
//* and the specified field-width as well as the value of the LC_NUMERIC locale*
//* environment variable. Examples are for the 'C' and U.S. English locale.    *
//* 1) Simple comma formatted output if specified field-width is sufficient.   *
//*    Examples:  345    654,345    782,654,345    4,294,967,295               *
//* 2) Compression of the value to fit a specified field width.                *
//*    Examples:  12.3K    999K    12.345M    1.234G    4.3G                   *
//*                                                                            *
//* Note on 'K', 'M', 'G' and 'T' suffixes: Because we are working with        *
//* decimal (base-10) values, compressed values are expressed as:              *
//*          kilobytes (x/1000)           NOT kibibytes (x/1024),              *
//*          megabytes (x/1000000)        NOT mibibytes (x/1024000) and        *
//*          gigabytes (x/1000000000),    NOT gibibytes (x/1024000000).        *
//*          terabytes (x/1000000000000), NOT tibibytes (x/1024000000000).     *
//* Please see NcDialog test application, 'Dialogw' for more examples.)        *
//*                                                                            *
//* Input  : iVal  : value to be formatted                                     *
//*                  Max supported value: 9.9999999999 terabytes==9.9999999999T*
//*          fWidth: field width (number of display columns)                   *
//*                  range: 1 to FI_MAX_FIELDWIDTH                             *
//*          lJust : (optional, false by default)                              *
//*                  if true, strip leading spaces to left-justify the value   *
//*                  in the field. (resulting string may be less than fWidth)  *
//*          sgnVal: (optional, false by default) [CURRENTLY IGNORED]          *
//*                  if true, then interpret 'iVal' as a signed number         *
//*          pnSign: (optional, false by default) [CURRENTLY IGNORED]          *
//*                  if 'sgnVal' == true AND pnSign == true, then prepend a    *
//*                  '+' or '-' sign to the value. The sign reduces precision  *
//*                  by one digit                                              *
//*                                                                            *
//* Returns: 'true' if successful                                              *
//*          'false' if field overflow (field will be filled with '#' chars)   *
//*                     note that oveflow can occur ONLY:                      *
//*                     if fWidth <= 3 OR if iVal >= 10. terabytes             *
//******************************************************************************
//* For 'short' and 'int' integers, we promote the 'iVal' parameter and call   *
//* the unsigned long int method.                                              *
//*                                                                            *
//* For 64-bit ('long long') integers, we break the unsigned long long int     *
//* value into two pieces and call the 32-bit method twice, and then           *
//* concatenate the result.                                                    *
//******************************************************************************

bool gString::formatInteger ( unsigned long iVal, short fWidth, bool lJust, 
                              bool sgnVal, bool pnSign )
{
   #if IFORMAT_EXTENSION != 0

   wchar_t wsrc[FI_BUFFER_SIZE] ;
   bool overflow = this->intFormat32 ( wsrc, iVal, fWidth, lJust, sgnVal, pnSign ) ;

   #else    // IFORMAT_EXTENSION disabled
   bool  overflow = true ;
   wchar_t wsrc[16] = { L"#" } ;
   #endif   // IFORMAT_EXTENSION

   //* Reinitialize all data members and convert *
   //* source text from wchar_t to UTF-8.        *
   this->Reinit() ;
   this->ConvertWidetoUTF8 ( wsrc ) ;

   return overflow ;

}  //* End formatInteger() *

bool gString::formatInteger ( unsigned long long iVal, short fWidth, bool lJust, 
                              bool sgnVal, bool pnSign )
{
   #if IFORMAT_EXTENSION != 0

   wchar_t wsrc[FI_BUFFER_SIZE] ;
   bool overflow = this->intFormat64 ( wsrc, iVal, fWidth, lJust, sgnVal, pnSign ) ;

   #else    // IFORMAT_EXTENSION disabled
   bool  overflow = true ;
   wchar_t wsrc[16] = { L"#" } ;
   #endif   // IFORMAT_EXTENSION

   //* Reinitialize all data members and convert *
   //* source text from wchar_t to UTF-8.        *
   this->Reinit() ;
   this->ConvertWidetoUTF8 ( wsrc ) ;

   return overflow ;

}  //* End formatInteger() *

bool gString::formatInteger ( unsigned int iVal, short fWidth, bool lJust, 
                              bool sgnVal, bool pnSign )
{
   return ( this->formatInteger( (unsigned long)iVal, fWidth, 
                                 lJust, sgnVal, pnSign ) ) ;
}  //* End formatInteger() *

bool gString::formatInteger ( unsigned short int iVal, short fWidth, bool lJust, 
                              bool sgnVal, bool pnSign )
{
   return ( this->formatInteger( (unsigned long)iVal, fWidth, 
                                 lJust, sgnVal, pnSign ) ) ;
}  //* End formatInteger() *

//*************************
//*       gstr            *
//*************************
//******************************************************************************
//* Return a pointer to the array of wchar_t (wide) characters.                *
//*                                                                            *
//* Input  : (by reference, initial value ignored)                             *
//*          receives number of characters in array, including null terminator *
//*                                                                            *
//* Returns: pointer to array of wchar_t characters                            *
//******************************************************************************

const wchar_t* gString::gstr ( short& charCount ) const
{
   charCount = this->gsChars ;   // characters in the array
   return ( this->gStr ) ;       // pointer to array

}  //* End gstr() *

//*************************
//*       ustr            *
//*************************
//******************************************************************************
//* Return a pointer to the array of UTF-8-encoded char characters.            *
//*                                                                            *
//* Input  : charCount:(by reference, initial value ignored)                   *
//*          receives number of characters in array, including null terminator *
//*        : byteCount:(by reference, initial value ignored)                   *
//*          receives number of data bytes in array                            *
//*                                                                            *
//* Returns: pointer to array of const char characters                         *
//******************************************************************************

const char* gString::ustr ( short& charCount, short& byteCount ) const 
{
   charCount = this->gsChars ;   // characters in the array
   byteCount = this->utfBytes ;  // bytes in the array
   return ( this->uStr ) ;       // pointer to array

}  //* End ustr() *

//*************************
//*     operator==        *
//*************************
//******************************************************************************
//* Comparison operator: Compares the text content of two gString objects.     *
//* Actual comparison is done on the objects' wchar_t character arrays.        *
//*                                                                            *
//*                                                                            *
//* Input  : (by reference) gString object containing string to be compared    *
//*                                                                            *
//* Returns: 'true' if the strings are identical, else 'false'                 *
//******************************************************************************

bool gString::operator== ( const gString& gs2 ) const
{

   short diff = this->compare ( gs2.gstr() ) ;
   return ( diff == ZERO ) ;

}  //* End operator==() *

//*************************
//*     operator!=        *
//*************************
//******************************************************************************
//* Comparison operator: Compares the text content of two gString objects.     *
//* Actual comparison is done on the objects' wchar_t character arrays.        *
//*                                                                            *
//* Input  : (by reference) gString object containing string to be compared    *
//*                                                                            *
//* Returns: 'true' if the strings are different, else 'false'                 *
//******************************************************************************

bool gString::operator!= ( const gString& gs2 ) const
{

   short diff = this->compare ( gs2.gstr() ) ;
   return ( diff != ZERO ) ;

}  //* End operator!=() *

//*************************
//*       compare         *
//*************************
//******************************************************************************
//* Compares the text content of the gString object with the specified text.   *
//* For both, actual comparison is done on  wchar_t character arrays.          *
//*                                                                            *
//* Input  : utf2 (UTF-8 string) OR wch2 (wchar_t string) to be compared       *
//*          charLimit: (optional, gsMAXCHARS by default)                      *
//*                     maximum number of characters to compare                *
//*          offset   : (optional, ZERO by default)                            *
//*                     If specified, equals the character offset into the     *
//*                     gString character array at which to begin comparison.  *
//*                                                                            *
//* Returns: using the rules of the 'wcsncmp' library function (see string.h): *
//*     ZERO, text data are identical                                          *
//*   > ZERO, first differing char of gString object is numerically larger.    *
//*   < ZERO, first differing char of gString object is numerically smaller.   *
//******************************************************************************

short gString::compare ( const wchar_t* wch2, short charLimit, short offset ) const
{

   return ( wcsncmp ( &this->gStr[offset], wch2, charLimit ) ) ;

}  //* End compare() *

short gString::compare ( const char* utf2, short charLimit, short offset ) const
{
   gString gs2( utf2 ) ;
   return ( this->compare ( gs2.gstr(), charLimit, offset ) ) ;

}  //* End compare() *

//*************************
//*        copy           *
//*************************
//******************************************************************************
//* Copy gString text to specified target buffer. (source data unchanged)      *
//*                                                                            *
//*                                                                            *
//* Input  : uTarget : pointer to target array to receive UTF-8-encoded text   *
//*          maxBytes: maximum number of bytes to copy (incl. NULL terminator) *
//*          maxCols : (optional, default == gsMAXCHARS*2)                     *
//*                    maximum number of display-columns to copy               *
//*                                                                            *
//* Returns: number of bytes copied (incl. NULL terminator)                    *
//******************************************************************************
//* Programmer's Note: If source data bytes greater than byte limit, copy      *
//* up to the whole UTF-8 character <= byte limit. This avoids creating an     *
//* invalid character in the target array.                                     *
//******************************************************************************

short gString::copy ( char* uTarget, short maxBytes, short maxCols ) const
{
   short copiedBytes = this->utfBytes ;   // return value

   //* If source data is within the specified limits, do the easy copy *
   if ( this->utfBytes <= maxBytes && this->gsCols <= maxCols )
   {
      for ( short i = ZERO ; i < this->utfBytes ; i++ )
         uTarget[i] = this->uStr[i] ;
   }
   //* Else, caller wants less that the whole string. *
   //* We preserve the original data in this object.  *
   else
   {
      gString gs = *this ;
      if ( gs.utfbytes() > maxBytes )
         gs.limitBytes( maxBytes ) ;
      if ( gs.gscols() > maxCols )
         gs.limitColumns( maxCols ) ;
      gs.copy( uTarget, maxBytes ) ;
      copiedBytes = gs.utfbytes() ;
   }
   return copiedBytes ;

}  //* End copy() *

//*************************
//*        copy           *
//*************************
//******************************************************************************
//* Copy gString text to specified target buffer. (source data unchanged)      *
//*                                                                            *
//*                                                                            *
//* Input  : wTarget : pointer to target array to receive wchar_t 'wide' text  *
//*          maxChars: maximum number of characters to copy (incl. NULL)       *
//*          maxCols : (optional, default == gsMAXCHARS*2)                     *
//*                    maximum number of display-columns to copy               *
//*                                                                            *
//* Returns: number of characters copied (incl. NULL terminator)               *
//******************************************************************************

short gString::copy ( wchar_t* wTarget, short maxChars, short maxCols ) const
{
   short copiedChars = this->gsChars ;    // return value

   //* If source data is within the specified limits, do the easy copy *
   if ( this->gsChars <= maxChars && this->gsCols <= maxCols )
   {
      for ( short i = ZERO ; i < this->gsChars ; i++ )
         wTarget[i] = this->gStr[i] ;
   }
   //* Else, caller wants less that the whole string. *
   //* We preserve the original data in this object.  *
   else
   {
      gString gs = *this ;
      // Note: limitCharacters() argument DOES NOT include the NULL terminator*
      if ( gs.gschars() > maxChars )
         gs.limitCharacters( (maxChars - 1) ) ;
      if ( gs.gscols() > maxCols )
         gs.limitColumns( maxCols ) ;
      gs.copy( wTarget, maxChars ) ;
      copiedChars = gs.gschars () ;

   }
   return copiedChars ;

}  //* End copy() *

//*************************
//*       append          *
//*************************
//******************************************************************************
//* Append text to existing gString text data up to a combined length of       *
//* gsMAXCHARS. Characters in excess of the maximum will not be appended.      *
//*                                                                            *
//* Input  : wPtr: pointer to array of wchar_t 'wide' text to be appended      *
//*           OR                                                               *
//*          uPtr: pointer to array of char UTF-8 text to be appended          *
//*           OR                                                               *
//*          wChar: a single, 'wide' character                                 *
//*                                                                            *
//* Returns: number of characters in resulting string (incl. NULL terminator)  *
//*          Note: if return == gsMAXCHARS, some data MAY HAVE BEEN discarded. *
//******************************************************************************

short gString::append ( const wchar_t* wPtr )
{
   if ( wPtr != NULL && *wPtr != NULLCHAR && (this->gsChars < (gsMAXCHARS - 1)) )
   {
      //* Append the new data to existing data *
      short srcIndex = ZERO,
            trgIndex = this->gsChars - 2 ;
      do
      {
         this->gStr[++trgIndex] = wPtr[srcIndex++] ;
      }
      while ( this->gStr[trgIndex] != NULLCHAR && trgIndex < (gsMAXCHARS - 1) ) ;
      this->gStr[trgIndex] = NULLCHAR ;   // be very sure that we are terminated
      this->gsChars = trgIndex + 1 ;
   
      //* Convert wide data to UTF-8 data and recalculate statistics *
      this->ConvertWidetoUTF8 () ;
   }
   return this->gsChars ;

}  //* End append() *

short gString::append ( const char* uPtr )
{
   gString  gs( uPtr ) ;
   return short(this->append( gs.gstr() )) ;

}  //* End append() *

short gString::append ( const wchar_t wChar )
{
   wchar_t wstr[] = { wChar, NULLCHAR } ;
   return short(this->append( wstr )) ;

}  //* End append() *

//*************************
//*       insert          *
//*************************
//******************************************************************************
//* Insert text into existing gString text data up to a combined length of     *
//* gsMAXCHARS. Characters in excess of the maximum will be truncated.         *
//*                                                                            *
//* Input  : wPtr  : pointer to array of wchar_t 'wide' text to be inserted    *
//*           OR                                                               *
//*          uPtr  : pointer to array of char UTF-8 text to be inserted        *
//*           OR                                                               *
//*          wChar : a single wchar_t 'wide' character                         *
//*          offset: (optional, ZERO by default)                               *
//*                  character offset at which to insert specified text into   *
//*                  existing text.                                            *
//*                  Note: if specified 'offset' > number of characters in     *
//*                        existing text, then acts like 'append' method.      *
//*                                                                            *
//* Returns: number of characters in resulting string (incl. NULL terminator)  *
//*          Note: if return == gsMAXCHARS, some data MAY HAVE BEEN discarded. *
//******************************************************************************
//* Programmer's Note: Caller's data could be of arbitrary length; however,    *
//* we have limited space, so some/all of existing data may be discarded, AND  *
//* some/all of new data may be truncated. This is the caller's problem, not   *
//* ours. If 'gsChars' == gsMAXCHARS, it is an indication to caller that some  *
//* data may have been discarded.                                              *
//*                                                                            *
//* Programmer's Note: We COULD HAVE offered a method for inserting one        *
//* single-byte character; however, we do not. UTF-8 characters are multibyte  *
//* by nature, so we would have to deny caller's request if the character      *
//* provided was not a valid, single-byte character. That would not be elegant.*
//******************************************************************************

short gString::insert ( const wchar_t* wPtr, short offset )
{
   if ( wPtr != NULL && *wPtr != NULLCHAR )
   {  //* If specified offset is beyond the end of *
      //* existing data, then do it the easy way.  *
      short nullIndex = (this->gsChars - 1) ;
      if ( offset >= nullIndex )
         this->append( wPtr ) ;

      //* Insert the new text at the specified position. *
      else
      {
         wchar_t trg[gsMAXCHARS + 10] ;      // temp work buffer
         short   srcIndex = ZERO,            // index into wPtr data
                 trgIndex = ZERO,            // index into trg buffer
                 gstIndex = ZERO ;           // index ino this->gStr[]

         //* If new data is to be inserted WITHIN the existing string, *
         //* write the 'before' text to the work buffer.               *
         if ( offset > ZERO )
         {
            for ( trgIndex = ZERO ; 
                  trgIndex < offset && trgIndex < nullIndex ; trgIndex++ )
               trg[trgIndex] = this->gStr[trgIndex] ;
            gstIndex = trgIndex ;

         }
         //* Append the new text to 'before'text. *
         while ( wPtr[srcIndex] != NULLCHAR && trgIndex < (gsMAXCHARS - 1) )
            trg[trgIndex++] = wPtr[srcIndex++] ;

         //* Append the 'after' text. *
         while ( trgIndex < (gsMAXCHARS - 1) && gstIndex < nullIndex )
            trg[trgIndex++] = this->gStr[gstIndex++] ;

         //* Be very sure that the string is terminated *
         trg[trgIndex] = NULLCHAR ;

         //* Copy the new string to our member storage and update statistics.*
         this->operator=( trg ) ;
      }
   }
   return this->gsChars ;

}  //* End insert() *

short gString::insert ( const char* uPtr, short offset )
{
   gString  gs( uPtr ) ;
   return short(this->insert( gs.gstr(), offset )) ;

}  //* End insert() *

short gString::insert ( wchar_t wChar, short offset )
{
   wchar_t wString[2] = { wChar, NULLCHAR } ;
   return short(this->insert( wString, offset )) ;

}  //* End insert() *

//*************************
//*      shiftChars       *
//*************************
//******************************************************************************
//* Shift text data by the specified number of characters.                     *
//*                                                                            *
//* Input  : shiftCount: < ZERO: shift data to the left, discarding the        *
//*                              specified number of characters from the       *
//*                              beginning of the array                        *
//*                      > ZERO: shift data to the right, padding the vacated  *
//*                              positions on the left with 'padChar'          *
//*                      ==ZERO: do nothing                                    *
//*          padChar   : (optional, SPACE character, 0x20 by default)          *
//*                      when shifting data to the right, use this character   *
//*                      to fill the vacated character positions               *
//* Returns: number of characters in adjusted array                            *
//******************************************************************************

short gString::shiftChars ( short shiftCount, wchar_t padChar )
{
   if ( shiftCount > ZERO )
   {
      if ( shiftCount < gsMAXCHARS )
      {
         short sIndex = this->gsChars - 1,      // source index
               tIndex = sIndex + shiftCount ;   // target index
         this->gsChars += shiftCount ;
         //* If we will shift part of the data out on the right *
         if ( this->gsChars > gsMAXCHARS )
         {
            sIndex -= this->gsChars - gsMAXCHARS + 1 ;
            this->gsChars = gsMAXCHARS ;
            tIndex = this->gsChars - 1 ;
            this->gStr[tIndex--] = NULLCHAR ;
         }
         while( sIndex >= ZERO )
            this->gStr[tIndex--] = this->gStr[sIndex--] ;
         while ( tIndex >= ZERO )
            this->gStr[tIndex--] = padChar ;
         this->ConvertWidetoUTF8 () ;  // update all status to match new data
      }
      else        // shift everything out
         this->Reinit () ;
   }
   else
   {
      shiftCount = ~shiftCount + 1 ;
      if ( shiftCount < this->gsChars )
      {
         short sIndex = shiftCount,    // source index
               tIndex = ZERO ;         // target index
         for ( ; this->gStr[sIndex] != NULLCHAR ; tIndex++ )
            this->gStr[tIndex] = this->gStr[sIndex++] ;
         this->gStr[tIndex++] = NULLCHAR ;
         this->gsChars = tIndex ;
         this->ConvertWidetoUTF8 () ;  // update all status to match new data
      }
      else        // shift everything out
         this->Reinit () ;
   }
   return this->gsChars ;

}  //* End shiftChars() *

//*************************
//*       shiftCols       *
//*************************
//******************************************************************************
//* Shift text data by the specified number of display columns.                *
//*                                                                            *
//* Input  : shiftCount: < ZERO: shift data to the left, discarding the        *
//*                              number of characters equivalent to the        *
//*                              specified number of display columns           *
//*                              NOTE: May discard one extra column if count   *
//*                              falls within a multi-column character.        *
//*                      > ZERO: shift data to the right, padding the vacated  *
//*                              positions on the left with 'padChar'          *
//*                      ==ZERO: do nothing                                    *
//*          padChar   : (optional, SPACE character, 0x20 by default)          *
//*                      when shifting data to the right, use this character   *
//*                      to fill the vacated column positions                  *
//*                      NOTE: Specify a one-column character ONLY as the      *
//*                      padding character. (multi-column characters ignored)  *
//* Returns: number of columns in adjusted array                               *
//******************************************************************************

short gString::shiftCols ( short shiftCount, wchar_t padChar )
{
   if ( shiftCount > ZERO )
   {
      if ( (wcwidth ( padChar )) != 1 )   // test width of padding character
         padChar = SPACE ;

      //* Max possible display columns == columns used + characters unused *
      short possibleCols = this->gsCols + (gsMAXCHARS - this->gsChars) ;
      if ( shiftCount < possibleCols )
      {
         short sIndex = this->gsChars - 1,      // source index
               tIndex = sIndex,                 // target index
               xIndex = ZERO,                   // terminal index
               xCols  = ZERO ;                  // column counter
         if ( shiftCount < this->gsCols )
         {
            while ( xCols < shiftCount && this->gStr[xIndex] != NULLCHAR )
               xCols += this->cWidth[xIndex++] ;
            xIndex += xCols - xIndex ;
         }
         else  // shift >= length of the array
            xIndex = shiftCount ;
         tIndex += xIndex ;
         this->gsChars += xIndex ;
         //* If we will shift part of the data out on the right *
         if ( this->gsChars > gsMAXCHARS )
         {
            sIndex -= this->gsChars - gsMAXCHARS + 1 ;
            this->gsChars = gsMAXCHARS ;
            tIndex = this->gsChars - 1 ;
            this->gStr[tIndex--] = NULLCHAR ;
         }
         while ( sIndex >= ZERO )
            this->gStr[tIndex--] = this->gStr[sIndex--] ;
         while ( tIndex >= ZERO )
            this->gStr[tIndex--] = padChar ;
         this->ConvertWidetoUTF8 () ;  // update all status to match new data
      }
      else        // shift everything out
         this->Reinit () ;
   }
   else if ( shiftCount < ZERO )
   {
      shiftCount = ~shiftCount + 1 ;
      if ( shiftCount < this->gsCols )
      {
         short xCols  = ZERO,          // number of columns for discard chars
               xIndex = ZERO,          // number of discard chars
               tIndex = ZERO ;         // target index
         while ( xCols < shiftCount && this->gStr[xIndex] != NULLCHAR )
            xCols += this->cWidth[xIndex++] ;
         for ( ; this->gStr[xIndex] != NULLCHAR ; tIndex++ )
            this->gStr[tIndex] = this->gStr[xIndex++] ;
         this->gStr[tIndex++] = NULLCHAR ;
         this->gsChars = tIndex ;
         this->ConvertWidetoUTF8 () ;  // update all status to match new data
      }
      else        // shift everything out
         this->Reinit () ;
   }
   return this->gsCols ;

}  //* End shiftCols() *

//*************************
//*       gschars         *
//*************************
//******************************************************************************
//* Returns the number of characters in the string including null terminator.  *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: number of characters in the string                                *
//******************************************************************************

short gString::gschars ( void ) const
{

   return this->gsChars ;

}  //* End gschars() *

//*************************
//*       gscols          *
//*************************
//******************************************************************************
//* Returns the number of columns required to display the string.              *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: number of columns needed to display the data                      *
//******************************************************************************

short gString::gscols ( void ) const
{

   return this->gsCols ;

}  //* End gscols() *

//*************************
//*        gscols         *
//*************************
//******************************************************************************
//* Returns a pointer to an array of column counts, one for each character     *
//* of the text data. Note that number of entries == number of characters.     *
//*                                                                            *
//* Input  : charCount (by reference, initial value ignored)                   *
//*           on return, contains number of characters in the string including *
//*           the null terminator.                                             *
//*                                                                            *
//* Returns: pointer to number of columns needed for display of each character *
//******************************************************************************

const short* gString::gscols ( short& charCount ) const
{

   charCount = this->gsChars ;
   return this->cWidth ;

}  //* End gscols() *

//*************************
//*      utfbytes         *
//*************************
//******************************************************************************
//* Returns the number of bytes in the UTF-8-encoded string.                   *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: number of bytes in UTF-8 string                                   *
//******************************************************************************

short gString::utfbytes ( void ) const
{

   return this->utfBytes ;

}  //* End utfbytes() *

//*************************
//*      isASCII          *
//*************************
//******************************************************************************
//* Scan the data to determine whether it is pure, 7-bit ASCII.                *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: 'true' if data are pure, 7-bit ASCII, else 'false'                *
//******************************************************************************

bool gString::isASCII ( void )
{
   this->isAscii = true ;
   if ( this->gsChars > ZERO )
   {
      for ( short i = ZERO ; this->uStr[i] != NULLCHAR ; i++ )
      {
         if ( this->uStr[i] < ZERO || this->uStr[i] > TILDE )
         {
            this->isAscii = false ;
            break ;
         }
      }
   }
   return this->isAscii ;

}  //* End isASCII() *

//*************************
//*     limitColumns      *
//*************************
//******************************************************************************
//* Truncate the existing data to no more than colCount display columns.       *
//*                                                                            *
//*                                                                            *
//* Input  : maximum number of display columns available for data display      *
//*                                                                            *
//* Returns: number of columns needed to display the adjusted data             *
//******************************************************************************

short gString::limitColumns ( short colCount )
{
   if ( colCount < ZERO || colCount > (gsMAXCHARS * 2) )
      colCount = (gsMAXCHARS * 2) ;
   this->ColumnCount ( colCount ) ;
   return this->gsCols ;

}  //* End limitColumns() *

//*************************
//*    limitCharacters    *
//*************************
//******************************************************************************
//* Truncate the existing data to no more than charCount display characters.   *
//*                                                                            *
//*                                                                            *
//* Input  : maximum number of characters allowed in formatted data            *
//*          (not including NULL)                                              *
//*                                                                            *
//* Returns: number of characters in the adjusted data (including NULL)        *
//******************************************************************************

short gString::limitCharacters ( short charCount )
{
   if ( charCount < ZERO || charCount >= gsMAXCHARS ) // range check
      charCount = gsMAXCHARS - 1 ;
   this->gStr[charCount] = NULLCHAR ;  // truncate the wchar_t array
   //* Recount the wide characters *
   for ( this->gsChars = ZERO ; this->gStr[this->gsChars] != NULLCHAR ; ++this->gsChars ) ;
   ++this->gsChars ;                   // count the NULLCHAR
   this->ConvertWidetoUTF8 () ;        // reconvert from wchar_t to UTF-8
   return this->gsChars ;

}  //* End limitCharacters() *

//*************************
//*  substringCharacters  *
//*************************
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************
// FOR FUTURE IMPLEMENTATION

//*************************
//*   substringColumns    *
//*************************
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************
// FOR FUTURE IMPLEMENTATION

//*************************
//*       insert          *
//*************************
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************
// FOR FUTURE IMPLEMENTATION

//*************************
//*  Get_gString_Version  *
//*************************
//******************************************************************************
//* Return a pointer to gString class version number string                    *
//*                                                                            *
//* Input  : none                                                              *
//*                                                                            *
//* Returns: pointer to version string                                         *
//******************************************************************************

const char* gString::Get_gString_Version ( void ) const
{

   return gStringVersion ;

}  //* End Get_gString_Version() *

//******************************************************************************
//*                         PROTECTED METHODS                                  *
//******************************************************************************
// None at this time.

//******************************************************************************
//*                           PRIVATE METHODS                                  *
//******************************************************************************

//*************************
//*  ConvertUTF8toWide    *
//*************************
//******************************************************************************
//* Convert a UTF-8-encoded string of up to gsMAXCHARS to a wchar_t            *
//* (wide character) string (null-terminated).                                 *
//* Saves a copy of the source data up to gsMAXBYTES (null-terminated)         *
//*                                                                            *
//* Input  : usrc : pointer to null-terminated UTF-8-encoded string            *
//*                 (includes ASCII strings as well as multi-byte characters)  *
//*                 (Assumes all class data members have been reinitialized)   *
//*          charLimit: (optional, gsMAXCHARS by default)                      *
//*                  maximum number of characters from source array to convert *
//*                                                                            *
//* Returns: nothing                                                           *
//*          On return, the following data members are initialized:            *
//*           gStr     : array of converted characters                         *
//*           gsChars  : number of converted characters                        *
//*                      (ZERO if conversion error - but see Bytewise_u2w())   *
//*           gsCols   : number of display columns needed to display gStr      *
//*           uStr     : UTF-8 source data, truncated if necessary             *
//*           utfBytes : number of bytes in (adjusted) source data             *
//******************************************************************************
//* Programmer's Note:                                                         *
//* The ISOC90 function, mbstowcs() is considered to be dangerous in that it   *
//* is not re-entrant or 're-startable'. For this reason, we use the           *
//* restartable version, mbsrtowcs(). This necessitates an understanding of    *
//* the mbstate_t structure.                                                   *
//*                                                                            *
//*  typedef struct                                                            *
//*  {                                                                         *
//*    int __count;                                                            *
//*    union                                                                   *
//*    {                                                                       *
//*      wint_t __wch;                                                         *
//*      char __wchb[4];                                                       *
//*    } __value;            /* Value so far.  */                              *
//*  } __mbstate_t;                                                            *
//*                                                                            *
//* On return from mbsrtowcs() this structure is updated with the current      *
//* 'state'. If NULLCHAR terminator was processed and there was no error, this *
//* is the 'initial' state.                                                    *
//*                     count == 0  ,   value == 0                             *
//* Otherwise an intermediate state is returned, reflecting what had happened  *
//* within the called function during the last character processed. This is    *
//* not particularly interesting to us.                                        *
//*  For a limited number of character conversions (NULLCHAR not processed):   *
//*   (only ASCII)      count == 0  ,   value == 0                             *
//*   (multi-byte)      count == 0  ,   value == some value > ZERO,            *
//*                       but not necessarily the most recent multi-byte       *
//*                       character processed. Unable to determine the meaning *
//*                       of the value without serious experimentation, which  *
//*                       I'm not willing to do at this time...)               *
//*  For an invalid multi-byte character encountered:                          *
//*                     count == 0  ,   value == 0  (for some random errors)   *
//*                                                                            *
//* What we want is to be sure that processing for each string begins in the   *
//* initial state and that any calls to the function by other threads will     *
//* not affect the processing in this one.                                     *
//*                                                                            *
//* Frankly, this whole alarmist scenario described in the man pages seems     *
//* bogus for any but the most esoteric of operations. Most character sets are *
//* stateless, and ALL UTF-8-encoded text is stateless. Although it is possible*
//* that the application is running in a non-UTF-8 environment, the NCurses/   *
//* NcWindow/NcDialog classes fully support only UTF-8 multi-byte encoding,    *
//* so they deserve whatever mess they've gotten themselves into...            *
//*                                                                            *
//* size_t mbstrowcs(wchar_t* DST, const char* SRC, size_t LEN, mbstate_t* PS);*
//* - returns number of wide characters converted OR -1 if invalid multi-byte  *
//*   char in source                                                           *
//* - DST is the target buffer to hold the wchar_t converted data              *
//* - SRC is the multi-byte source string (for us, always UTF-8)               *
//*   Note: according to the man page:                                         *
//*      "If DST is not null,`mbsrtowcs' stores in the pointer pointed to by   *
//*       SRC either a null pointer (if the NUL byte in the input string was   *
//*       reached) or the address of the byte following the last converted     *
//*       multibyte character."                                                *
//*   Frankly, we didn't know what the hell this meant, but in fact *SRC is    *
//*   modified by the call. This is beyond reasoning, but we protect the       *
//*   our input parameter 'const char* usrc' from modification, so we can      *
//*   use it later to save a copy of caller's multi-byte data string.          *
//* - LEN is max characters to convert                                         *
//* - PS is the 'processing state'                                             *
//*                                                                            *
//*                                                                            *
//*                                                                            *
//******************************************************************************

void gString::ConvertUTF8toWide ( const char* usrc, short charLimit )
{
   if ( usrc != NULL && *usrc != NULLCHAR )
   {
      //* Convert UTF-8 source to wchar_t target *
      mbstate_t   ps = { ZERO, { ZERO } } ;
      const char*  us = usrc ;   // protect the method parameter from modification
      const char** srcPtr = &us ;// see note above
      if ( (this->gsChars = mbsrtowcs ( this->gStr, srcPtr, charLimit, &ps )) > ZERO )
      {
         //* It's possible that no null terminator was appended during         *
         //* conversion, so do it now. Even if it was appended, it wasn't      *
         //* counted in the conversion total, so count it now.                 *
         //* If gsChars < charLimit, NULLCHAR was converted but not counted,   *
         //* so count it.                                                      *
         if ( this->gsChars < charLimit )
         {
            ++this->gsChars ;
   
            //* Copy the source to our local buffer.                           *
            //* Because the entire string was converted, simply copy it.       *
            short i = -1 ;
            do
            {
               ++i ;
               this->uStr[i] = usrc[i] ;
            }
            while ( this->uStr[i] != NULLCHAR ) ;
            this->utfBytes = ++i ;
            this->ColumnCount () ;     // count the resulting display columns
         }
         //* Else, gsChars == charLimit, NULLCHAR may or may not have been     *
         //* converted. If terminator not converted, add the null terminator   *
         //* and count it. Else, just count it.                                *
         else
         {  //* If buffer is full, check for terminator. If not present, add   *
            //* it, but with no change to count. If already present, just      *
            //* count it.                                                      *
            if ( charLimit == gsMAXCHARS )
            {
               if ( this->gStr[gsMAXCHARS-1] != NULLCHAR )
                  this->gStr[gsMAXCHARS-1] = NULLCHAR ;
               else
                  ++this->gsChars ;
            }
            //* Else conversion of characters was explicitly limited. If null  *
            //* terminator is not present, add it now. If already present,     *
            //* just count it.                                                 *
            else
            {
               if ( this->gStr[charLimit-1] != NULLCHAR )
                  this->gStr[charLimit] = NULLCHAR ;
               ++this->gsChars ;
            }
   
            //* Copy the source to our local buffer.                           *
            //* Because the string was truncated, we cannot be sure how many   *
            //* bytes to copy from the source string, so we re-convert to      *
            //* UTF-8 to get the number of bytes. We 'know' results will fit   *
            //* in our buffer.                                                 *
            this->ConvertWidetoUTF8 () ;
         }
      }
      else                             // conversion error
      {  //* If a conversion error, convert as much of the string as possible. *
         this->Bytewise_u2w ( usrc, charLimit ) ;
         this->ColumnCount () ;        // count the resulting display columns
      }
   }

}  //* End ConvertUTF8toWide() *

//*************************
//*     Bytewise_u2w      *
//*************************
//******************************************************************************
//* Convert the UTF-8 data in usrc to wchar_t data and store it in gStr.       *
//* This method is called ONLY when there is an error in conversion of the     *
//* string as a whole (see ConvertUTF8toWide()).                               *
//* Convert one character at a time until the null terminator has been         *
//* converted, OR a conversion error is encountered.                           *
//* Both gStr and uStr will be null-terminated on return.                      *
//*                                                                            *
//* Input  : usrc : pointer to null-terminated UTF-8-encoded string            *
//*                 (includes ASCII strings as well as multi-byte characters)  *
//*                 (Assumes all class data members have been reinitialized)   *
//*          charLimit: (optional, gsMAXCHARS by default)                      *
//*                  maximum number of characters from source array to convert *
//*                                                                            *
//* Returns: nothing                                                           *
//*          On return, the following data members are initialized:            *
//*           gStr     : array of converted characters up to conversion error  *
//*                      (all valid, and null-terminated)                      *
//*           gsChars  : number of characters successfully converted           *
//*                      includng null terminator                              *
//*           gsCols   : number of display columns needed to display gStr      *
//*           uStr     : truncated array of successfully-converted characters  *
//*                      (all valid, and null-terminated)                      *
//*           utfBytes : number of bytes in remaining character codes          *
//******************************************************************************
//* Note: mbrtowc() returns one of the following:                              *
//*       >0 == bytes in converted multi-byte character                        *
//*        0 == NULLCHAR converted                                             *
//*       -1 == invalid multi-byte sequence                                    *
//*       -2 == multi-byte char is incomplete at maxBPC (not very likely)      *
//*                                                                            *
//******************************************************************************
// -- Function: size_t mbrtowc (wchar_t *restrict PWC, const char
//          *restrict S, size_t N, mbstate_t *restrict PS)
//     The `mbrtowc' function ("multibyte restartable to wide character")
//     converts the next multibyte character in the string pointed to by
//     S into a wide character and stores it in the wide character string
//     pointed to by PWC.  The conversion is performed according to the
//     locale currently selected for the `LC_CTYPE' category.  If the
//     conversion for the character set used in the locale requires a
//     state, the multibyte string is interpreted in the state
//     represented by the object pointed to by PS.  If PS is a null
//     pointer, a static, internal state variable used only by the
//     `mbrtowc' function is used.
//
//     If the next multibyte character corresponds to the NUL wide
//     character, the return value of the function is 0 and the state
//     object is afterwards in the initial state.  If the next N or fewer
//     bytes form a correct multibyte character, the return value is the
//     number of bytes starting from S that form the multibyte character.
//     The conversion state is updated according to the bytes consumed in
//     the conversion.  In both cases the wide character (either the
//     `L'\0'' or the one found in the conversion) is stored in the
//     string pointed to by PWC if PWC is not null.
//
//     If the first N bytes of the multibyte string possibly form a valid
//     multibyte character but there are more than N bytes needed to
//     complete it, the return value of the function is `(size_t) -2' and
//     no value is stored.  Please note that this can happen even if N
//     has a value greater than or equal to `MB_CUR_MAX' since the input
//     might contain redundant shift sequences.
//
//     If the first `n' bytes of the multibyte string cannot possibly form
//     a valid multibyte character, no value is stored, the global
//     variable `errno' is set to the value `EILSEQ', and the function
//     returns `(size_t) -1'.  The conversion state is afterwards
//     undefined.
// 

void gString::Bytewise_u2w ( const char* usrc, short charLimit )
{
   //* Reinitialize all data members *
   this->Reinit() ;

   wchar_t  wtmp ;               // converted character
   short uBytes,                 // UTF-8 bytes for converted character
         wIndex = ZERO,          // index into gStr
         uIndex = ZERO ;         // index into uStr and usrc
   bool  done = false ;
   while ( ! done && wIndex < charLimit )
   {
      mbstate_t   ps = { ZERO, { ZERO } } ;// 'state' structure (initial state)
      uBytes = mbrtowc ( &wtmp, &usrc[uIndex], maxBPC, &ps ) ;

      //* If valid (non-null) wchar_t character stored in wtmp *
      if ( uBytes > ZERO )
      {
         this->gStr[wIndex++] = wtmp ;
         for ( short i = ZERO ; i < uBytes ; i++ ) // store multi-byte character
         {
            this->uStr[uIndex] = usrc[uIndex] ;
            ++uIndex ;
         }
      }
      //* If null terminator has just been converted OR                  *
      //* if no multi-byte sequence does not represent a valid character *
      else
      {
         this->uStr[uIndex] = NULLCHAR ;  // terminate both strings
         this->gStr[wIndex] = NULLCHAR ;
         this->gsChars = wIndex + 1 ;     // number of valid characters
         this->utfBytes = uIndex + 1 ;    // number of valid UTF-8 bytes stored
         done = true ;
      }
   }

}  //* End Bytewise_u2w() *

//*************************
//*  ConvertWidetoUTF8    *
//*************************
//******************************************************************************
//* Convert the wchar_t string in gStr to a UTF-8-encoded, multi-byte character*
//* string.                                                                    *
//*                                                                            *
//* Input  : wsrc: (optional, NULL pointer by default, and will operate on     *
//*                 existing data in gStr[] member.)                           *
//*                If non-NULL, points to null-terminated array of wchar_t     *
//*                 source characters.                                         *
//*          charLimit: (optional, gsMAXCHARS by default)                      *
//*                  maximum number of characters from source array to convert *
//*                                                                            *
//* Returns: nothing                                                           *
//*          On return, all data members are reinitialied with the possible    *
//*           exception of gStr[] and gsChars which MAY remain unchanged.      *
//******************************************************************************

void gString::ConvertWidetoUTF8 ( const wchar_t* wsrc, short charLimit )
{
   //* Text data will change, so reset everything but gStr[] and gsChars *
   this->Reinit ( false ) ;

   //* If external data specified, copy it to gStr data member *
   if ( wsrc != NULL )
   {
      short i = -1 ;
      do
      {
         if ( ++i < charLimit )     // index wchar_t storage
            this->gStr[i] = wsrc[i] ;  // store the character

         //* Target buffer full or explicit character limit reached *
         else
         {
            if ( charLimit == gsMAXCHARS )   // if buffer is full
               i = gsMAXCHARS - 1 ;          // max buffer index (position for NULL)
               // else explicit limit (non-NULL character count) reached
            this->gStr[i] = NULLCHAR ; // either way, terminate the string
         }
      }
      while ( this->gStr[i] != NULLCHAR ) ;
      this->gsChars = i + 1 ;       // character count incl. NULL
   }

   //* If we have a non_NULL string, convert it *
   if ( *this->gStr != NULLCHAR )
   {
      //* Convert wchar_t string to UTF-8 format *
      short maxbytes = gsMAXBYTES - 2 ;
      char  tmpstr[gsMAXBYTES] ;
      short tmpbytes ;
      mbstate_t   ps = { ZERO, { ZERO } } ;
      const wchar_t*  ws = this->gStr ;// protect buffer pointer from modification
      const wchar_t** srcPtr = &ws ;   // see note in ConvertUTF8toWide()
      if ( (tmpbytes = wcsrtombs ( tmpstr, srcPtr, maxbytes, &ps )) > ZERO )
      {
         //* It's possible that no null terminator was appended during         *
         //* conversion, so do it now. Even if it was appended, it wasn't      *
         //* counted in the  conversion total, so count it now.                *
         //* If tmpbytes < maxbytes, NULLCHAR was converted but not counted,   *
         //* so count it.                                                      *
         if ( tmpbytes < maxbytes )
            ++tmpbytes ;
         //* Else, tmpbytes == maxbytes, NULLCHAR may or may not have been     *
         //* converted. If it was, just count it, if not, add null char.       *
         else
         {
            if ( tmpstr[tmpbytes-1] != NULLCHAR )
               tmpstr[tmpbytes] = NULLCHAR ;
            ++tmpbytes ;
         }
   
         //* Copy converted string to uStr and update counters *
         short i = -1 ;
         do
         {
            ++i ;
            this->uStr[i] = tmpstr[i] ;
         }
         while ( this->uStr[i] != NULLCHAR ) ;
         this->utfBytes = tmpbytes ;
      }
      else                             // conversion error (or NULL string)
      {  //* If a conversion error, convert as much of the string as possible. *
         this->Bytewise_w2u ( charLimit ) ;
      }
      this->ColumnCount () ;        // count the resulting display columns
   }

}  //* End ConvertWidetoUTF8() *

//*************************
//*     Bytewise_w2u      *
//*************************
//******************************************************************************
//* Convert the wchar_t data in gStr to UTF-8 data and store it in uStr.       *
//* This method is called ONLY when there is an error in conversion of the     *
//* string as a whole (see ConvertWidetoUTF8()).                               *
//* Convert one character at a time until the null terminator has been         *
//* converted, OR a conversion error is encountered.                           *
//* Both gStr and uStr will be null-terminated on return.                      *
//*                                                                            *
//*                                                                            *
//* Input  : charLimit : maximum number of characters to convert.              *
//*                                                                            *
//* Returns: nothing                                                           *
//*          On return, the following data members are initialized:            *
//*           uStr     : array of converted characters up to conversion error  *
//*                      (all valid, and null-terminated)                      *
//*           utfBytes : number of bytes in converted array                    *
//*           gStr     : truncated array of successfully-converted characters  *
//*                      (all valid, and null-terminated)                      *
//*           gsChars  : number of remaining characters, incl. null terminator *
//******************************************************************************
//* Programmer's Note: It is unlikely that this method will ever be called.    *
//* We have verified that the method correctly converts the entire string,     *
//* terminating at NULLCHAR or any invalid wchar_t character.                  *
//* However, only wchar_t values with bit 31 set induce an error return from   *
//* either the  wcsrtombs() and wcrtomb() library functions, so an application *
//* would have to work really hard to screw up that badly.                     *
//******************************************************************************

void gString::Bytewise_w2u ( short charLimit )
{
   //* Text data will change, so reset everything but gStr[] and gsChars *
   this->Reinit ( false ) ;

   //* If we have a non_NULL string, convert it *
   if ( *this->gStr != NULLCHAR )
   {
      short maxbytes = gsMAXBYTES - 2, // output buffer limit
            uBytes,                    // UTF-8 bytes for converted character
            wIndex = ZERO,             // index into gStr
            uIndex = ZERO ;            // index into uStr
      bool  done = false ;
      while ( ! done && wIndex < charLimit )
      {
         mbstate_t   ps = { ZERO, { ZERO } } ;// 'state' structure (initial state)
         uBytes = wcrtomb ( &this->uStr[uIndex], this->gStr[wIndex], &ps ) ;
         //* If null terminator has just been converted OR         *
         //* if no multi-byte representation for this character OR *
         //* if output buffer is full (rather unlikely).           *
         if ( this->gStr[wIndex] == NULLCHAR || uBytes < ZERO 
             || ((uIndex + uBytes) >= maxbytes) )
         {
            if ( (uIndex + uBytes) == maxbytes )
            {
               ++wIndex ;
               uIndex += uBytes ;
            }
            this->uStr[uIndex] = NULLCHAR ;  // terminate both strings
            this->gStr[wIndex] = NULLCHAR ;
            this->gsChars = wIndex + 1 ;     // number of valid characters
            this->utfBytes = uIndex + 1 ;    // number of valid UTF-8 bytes stored
            done = true ;
         }
         //* Valid (non-null) multi-byte character stored *
         else
         {
            ++wIndex ;
            uIndex += uBytes ;
         }
      }
   }   

}  //* End Bytewise_w2u() *

//*************************
//*   Convert_gsForm2gs   *
//*************************
//******************************************************************************
//* Convert gsForm-class data to gString data.                                 *
//*                                                                            *
//*                                                                            *
//* Input  : gsf : an initialized gsForm class object containing formatting    *
//*                specification string and pointers to data values            *
//*          wsrc: pointer to caller's buffer to hold formatted                *
//*                wchar_t text data                                           *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//* Programmer's Note:                                                         *
//* - swprintf() returns the number of characters written, (excluding the null *
//*   terminator). If a value < ZERO is returned, then the output was          *
//*   truncated due to lack of output-buffer space.                            *
//* - Because we call swprintf() with only one format specification at a time, *
//*   if return value < ZERO, it means that the formatting operation was not   *
//*   performed at all, and that the output string has not been modified by    *
//*   the library call.                                                        *
//* - In this case, caller may be surprised to find that the output was        *
//*   truncated although the buffer was not completely filled. Sorry, can't    *
//*   be helped.                                                               *
//* - Possible future enhancement: If the failed formatting operation was for  *
//*   a string, we COULD copy as much of the string as will fit; however, at   *
//*   this time we feel it's more trouble than it's worth. 21 Oct. 2011        *
//*                                                                            *
//******************************************************************************

void gString::Convert_gsForm2gs ( const gsForm& gsf, wchar_t wsrc[gsMAXCHARS] )
{
const short fmtSPACE = 32 ;
wchar_t  fmt[fmtSPACE] ;         // holds format specification for swprintf() call
int      swret = ZERO ;          // if swprintf returns < zero, buffer is full
short    wsrci  = ZERO,          // index into wsrc[]
         fmti,                   // index into fmt[]
         fsi    = ZERO,          // index into gsf.fmtSpec[]
         pCount = ZERO,          // number of conversion symbols found
         wsrcSpace = gsMAXCHARS - 1 ;// remaining free space in wsrc buffer


   *wsrc = NULLCHAR ;            // start with an empty output buffer

   while ( (gsf.fmtSpec[fsi] != NULLCHAR) && (pCount < gsFormMAXARGS) && (wsrcSpace > 1) )
   {
      //* Scan for next conversion symbol, copying literal data as-is *
      if ( gsf.fmtSpec[fsi] != '%' )
      {
         wsrc[wsrci++] = gsf.fmtSpec[fsi++] ;
         wsrc[wsrci] = NULLCHAR ;   // keep string terminated at all times
         --wsrcSpace ;
      }
      else
      {  //* Conversion symbol found. Isolate the conversion *
         //* specification for this argument.                *
         fmti = ZERO ;
         fmt[fmti++] = gsf.fmtSpec[fsi++] ;    // save conversion symbol
         do
         {
            if (   (gsf.fmtSpec[fsi] == 's') || (gsf.fmtSpec[fsi] == 'S')
                || (gsf.fmtSpec[fsi] == 'c') || (gsf.fmtSpec[fsi] == 'C')
                || (gsf.fmtSpec[fsi] == 'd') || (gsf.fmtSpec[fsi] == 'i')
                || (gsf.fmtSpec[fsi] == 'x') || (gsf.fmtSpec[fsi] == 'X')
                || (gsf.fmtSpec[fsi] == 'u') || (gsf.fmtSpec[fsi] == 'o')
                || (gsf.fmtSpec[fsi] == 'f') || (gsf.fmtSpec[fsi] == 'p')
                || (gsf.fmtSpec[fsi] == 'e') || (gsf.fmtSpec[fsi] == 'E')
                || (gsf.fmtSpec[fsi] == 'g') || (gsf.fmtSpec[fsi] == 'G')
                || (gsf.fmtSpec[fsi] == 'a') || (gsf.fmtSpec[fsi] == 'A')
                || (gsf.fmtSpec[fsi] == 'n') || (gsf.fmtSpec[fsi] == 'm')
               )
            {
               //* Determine the data type for this conversion *
               char convType ;         // data type for converting argument
               fmt[fmti++] = convType = gsf.fmtSpec[fsi++] ; // save the data type
               fmt[fmti] = NULLCHAR ;  // terminate the specification string
               fmti -= 2 ;             // reference 'long' or 'short' modifier, if any
               if ( convType == 's' && fmt[fmti] == 'l' )
                  convType = 'S' ;     // wide character string
               else if ( convType == 'c' && fmt[fmti] == 'l' )
                  convType = 'C' ;     // wide character
               else if (   convType == 'd' || convType == 'i' || convType == 'x' 
                        || convType == 'X' || convType == 'u' || convType == 'o' )
               {  //* Determine the 'sizeof' flag for integer source *
                  if ( fmt[fmti] == 'h' )
                     convType = 'd' ;           // short int size
                  else if ( fmt[fmti] == 'q' || fmt[fmti] == 'L' 
                           || (fmti >= 2 && fmt[fmti] == 'l' && fmt[fmti-1] == 'l') )
                     convType = 'q' ;           // quad int (64-bit) size
                  else if ( fmt[fmti] == 'l' )
                     convType = 'D' ;           // long int size
                  else
                     convType = 'i' ;           // 'int' size
               }
               else if (   convType == 'f' 
                        || convType == 'e' || convType == 'E' 
                        || convType == 'g' || convType == 'G'
                        || convType == 'a' || convType == 'A'
                       )
               {
                  if ( fmt[fmti] == 'l' )
                     convType = 'F' ;
                  else if ( fmt[fmti] == 'L' )
                     convType = 'Q' ;
                  else
                     convType = 'f' ;
               }
               else if ( convType == 'n' )
               {
                  if ( fmt[fmti] == 'h' )
                     convType = 'n' ;
                  else
                     convType = 'N' ;
               }

               //* Create formatted output for this argument *
               swret = ZERO ; // initialize swprintf return value
                              // (in case it isn't called)
               if ( convType == 's' )
               {
                  // (see notes in composeFieldwidthBugFix())
                  if ( fmt[1] == 's' )       // no width specification
                  {
                     swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                    (const char*)gsf.argList[pCount] ) ;
                  }
                  else
                  {
                     gString gs( (const char*)gsf.argList[pCount] ) ;
                     swret = composeFieldwidthBugFix ( convType, &wsrc[wsrci], 
                                                       wsrcSpace, fmt, gs ) ;
                  }
               }
               else if ( convType == 'S' )
               {
                  if ( fmt[1] == 'S' || fmt[1] == 'l' )// no width specification
                  {
                     swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                    (const wchar_t*)gsf.argList[pCount] ) ;
                  }
                  else
                  {
                     gString gs( (const wchar_t*)gsf.argList[pCount] ) ;
                     swret = composeFieldwidthBugFix ( convType, &wsrc[wsrci], 
                                                       wsrcSpace, fmt, gs ) ;
                  }
               }
               else if ( convType == 'c' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 *((const char*)(gsf.argList[pCount])) ) ;
               else if ( convType == 'C' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 *((const wchar_t*)(gsf.argList[pCount])) ) ;
               else if ( convType == 'd' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 *((short*)(gsf.argList[pCount])) ) ;
               else if ( convType == 'i' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 *((int*)(gsf.argList[pCount])) ) ;
               else if ( convType == 'D' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 *((long int*)(gsf.argList[pCount])) ) ;
               else if ( convType == 'q' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 *((long long int*)(gsf.argList[pCount])) ) ;
               else if ( convType == 'f' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 *((float*)(gsf.argList[pCount])) ) ;
               else if ( convType == 'F' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 *((double*)(gsf.argList[pCount])) ) ;
               else if ( convType == 'Q' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 *((long double*)(gsf.argList[pCount])) ) ;
               else if ( convType == 'p' )
               {
                  // NOTE: There is a bug in swprintf() which renders a NULL
                  //       pointer as "(" rather than as "(nil)".
                  if ( gsf.argList[pCount] != NULL )
                     swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt,
                                        gsf.argList[pCount] ) ;
                  else
                     swret = swprintf ( &wsrc[wsrci], wsrcSpace, L"%S", L"(nil)" ) ;
               }
               else if ( convType == 'm' )
                  swret = swprintf ( &wsrc[wsrci], wsrcSpace, fmt, 
                                 gsf.argList[pCount] ) ;
               else if ( convType == 'n' )
               {
                  *((short*)(gsf.argList[pCount])) = wsrci ;
                  wsrc[wsrci] = NULLCHAR ; // produces no output to formatted string
                  swret = ZERO ;
               }
               else if ( convType == 'N' )
               {
                  *((int*)(gsf.argList[pCount])) = (int)wsrci ;
                  wsrc[wsrci] = NULLCHAR ; // produces no output to formatted string
               }
               else  // unsupported data types 'n' and unknown types
               { 
                  wsrc[wsrci++] = '#' ;
                  wsrc[wsrci] = NULLCHAR ;
               }

               //* Find end of string (see important note in method header) *
               if ( swret >= ZERO )
               {
                  while ( wsrc[wsrci] != NULLCHAR )
                     ++wsrci ;
               }
               else  // buffer full
               {
                  wsrc[gsMAXCHARS - 1] = NULLCHAR ;
                  wsrcSpace = ZERO ;
               }
               ++pCount ;        // reference next argument
               break ;           // finished with this conversion specification
            }  // if()

            #if BINARY_EXTENSION != 0
            else if ( (gsf.fmtSpec[fsi] == 'b') || (gsf.fmtSpec[fsi] == 'B') )
            {  //* Binary output extension to swprintf *
               //* (see notes in binaryFormat method)  *
               //* Determine the data type and modifiers for this conversion.*
               fmt[fmti++] = gsf.fmtSpec[fsi] ; // save the data type
               fmt[fmti] = NULLCHAR ;  // terminate the specification string
               fmti -= 2 ;             // reference 'long' or 'short' modifier, if any

               long long int intData ;
               wchar_t convBuff[100],  // receives formatted data
                       convType = gsf.fmtSpec[fsi++], // indicator: 'b' or 'B'
                       sepType = L'.' ;// default group seperator character
               short dataType,         // number of bytes in source integer
                     grpType = 4,      // grouping: by-4 or by-8
                     i = 1 ;
               bool appendType = true ;// assume append indicator (or none)

               //* Determine whether alternate seperation character specified.*
               if ( fmt[i] == '#' )    ++i ;
               else if ( fmt[i] == '-' && fmt[i+1] == '#' )   i += 2 ;
               for ( ; fmt[i] != NULLCHAR ; i++ )
               {
                  if ( ((fmt[i] >= SPACE && fmt[i] <= '/') && fmt[i] != '.') || 
                       (fmt[i] >= ':' && fmt[i] <= '@') || (fmt[i] == '[') || 
                       (fmt[i] >= ']' && fmt[i] <= '`') || 
                       (fmt[i] >= '{' && fmt[i] <= '~') )
                  { sepType = fmt[i] ; break ; }
               }

               //* Determine whether group-of-8 spacing specified.*
               //* ".8" (group-of-4 spacing is the default)       *
               for ( i = ZERO ; fmt[i] != NULLCHAR ; i++ )
               { if ( fmt[i] == '.' && fmt[i+1] == '8' ) { grpType = 8 ; break ; } }

               //* Initialize the data, data type and 'appendType' flag. *
               if ( fmt[fmti] == 'h' )
               {
                  if ( fmti >= 2 && fmt[fmti-1] == 'h' )
                  {
                     char x = *((char*)(gsf.argList[pCount])) ;
                     intData = (long long int)x ;
                     dataType = sizeof(char) ;     // byte int source
                  }
                  else
                  {
                     short x = *((short*)(gsf.argList[pCount])) ;
                     intData = (long long int)x ;
                     dataType = sizeof(short int) ;// short int source
                  }
               }
               else if ( fmt[fmti] == 'q' || fmt[fmti] == 'L' 
                        || (fmti >= 2 && fmt[fmti] == 'l' && fmt[fmti-1] == 'l') )
               {
                  intData = *((long long int*)(gsf.argList[pCount])) ;
                  dataType = sizeof(long long int) ;
               }
               else if ( fmt[fmti] == 'l' )
               {
                  long x = *((long*)(gsf.argList[pCount])) ;
                  intData = (long long int)x ;
                  dataType = sizeof(long int) ; // long int source
               }
               else
               {
                  int x = *((int*)(gsf.argList[pCount])) ;
                  intData = (long long int)x ;
                  dataType = sizeof(int) ;   // int source
               }
               if ( fmt[1] == '#' )    // prepend 'b' or 'B'
                  appendType = false ;
               else if ( fmt[1] != '-' || fmt[2] != '#' )
                  convType = NULLCHAR ;
               //* Format the data *
               this->binaryFormat ( convBuff, convType, sepType, intData, 
                                    dataType, grpType, appendType ) ;
               //* Add the formatted data to the output *
               for ( short i = ZERO ; convBuff[i] != NULLCHAR ; i++ )
               {
                  if ( wsrci < (gsMAXCHARS - 1) )
                     wsrc[wsrci++] = convBuff[i] ;
                  else
                  {
                     wsrci = gsMAXCHARS - 1 ;
                     break ;
                  }
               }
               wsrc[wsrci] = NULLCHAR ;
               ++pCount ;        // reference next argument
               break ;
            }
            #endif   // BINARY_EXTENSION

            else if ( gsf.fmtSpec[fsi] == '%' )
            {  //* Copy literal '%' to output string *
               wsrc[wsrci++] = gsf.fmtSpec[fsi++] ;
               wsrc[wsrci] = NULLCHAR ;
               break ;
            }
            else if ( gsf.fmtSpec[fsi] == NULLCHAR )
            {  //* Unsupported data type encountered *
               wsrc[wsrci++] = '#' ;
               wsrc[wsrci] = NULLCHAR ;
               break ;
            }
            else
            {  //* Copy specification modifier to specification string *
               fmt[fmti++] = gsf.fmtSpec[fsi++] ;
            }
         }
         while ( gsf.fmtSpec[fsi] != NULLCHAR ) ;

         //* Recalculate remaining free space in target buffer *
         wsrcSpace = gsMAXCHARS - wsrci - 1 ;
      }  // else
   }  // while()

   //* Be absolutely sure the string is terminated *
   if ( wsrci >= gsMAXCHARS )
      wsrci = gsMAXCHARS - 1 ;
   wsrc[wsrci] = NULLCHAR ;

}  //* End Convert_gsForm2gs() *

//*************************
//*     limitBytes        *
//*************************
//******************************************************************************
//* Truncate the existing data to no more than maxBytes bytes.                 *
//* This method is used internally by the copy() method to prevent overflow    *
//* of caller's buffer. This method IS NOT publically available.               *
//* The actual number of bytes in the truncated string is determined by the    *
//* number of whole-character byte sequences <= maxBytes. This avoids a        *
//* partial character sequence at the end of the string. The NULLCHAR          *
//* terminating character is included in the number of bytes retained.         *
//*                                                                            *
//* Input  : maximum number of bytes to be retained                            *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void gString::limitBytes ( short maxBytes )
{
   if ( maxBytes > 1 && maxBytes < this->utfBytes )
   {
      short bCount = ZERO,       // number of bytes retained
            btmp ;               // bytes in an individual character-byte sequence

      --maxBytes ;               // leave room for NULLCHAR terminator
      while ( bCount < maxBytes )
      {
         btmp = mbrlen ( &this->uStr[bCount], maxBPC, NULL ) ;
         if ( (bCount + btmp) <= maxBytes )
            bCount += btmp ;
         else
            break ;
      }
      char  uSource[gsMAXBYTES] ;  // temporary buffer
      for ( short i = ZERO ; i < bCount ; i++ )
         uSource[i] = this->uStr[i] ;
      uSource[bCount++] = NULLCHAR ;
      this->Reinit () ;       // reinitialize all data members
      this->ConvertUTF8toWide ( uSource ) ;
   }

}  //* End limitBytes() *

//*************************
//*      ColumnCount      *
//*************************
//******************************************************************************
//* Count the number of display columns required to display the data.          *
//* Optionally, truncate the data to limit the number of columns required.     *
//*                                                                            *
//* Input  : maxCols (optional, default==(gsMAXCHARS * 2))                     *
//*                  if specified, sets maximum number of display-output       *
//*                  columns allowed in string.                                *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************

void gString::ColumnCount ( short maxCols )
{
   short colCount ;
   this->gsCols = ZERO ;                  // reset column count
   for ( short i = ZERO ; i < this->gsChars ; i++ )
   {  //* Get width in display columns for the character *
      if ( (colCount = wcwidth ( this->gStr[i] )) >= ZERO )   // if no error
      {
         this->cWidth[i] = colCount ;        // save column width to tracking array
         if ( colCount > ZERO )
         {
            this->gsCols += colCount ;       // add char columns to running total
            if ( this->gsCols > maxCols )    // if max column count reached
            {
               this->gStr[i] = NULLCHAR ;    // discard most recent char examined
               this->cWidth[i] = ZERO ;
               this->gsCols -= colCount ;
               this->gsChars = i + 1 ;       // new char count (incl. null)
               //* This call will trash our new column data, so protect it. *
               short tempgsCols = this->gsCols,
                     tempcWidth = this->cWidth[ZERO] ;
               this->ConvertWidetoUTF8 () ;  // uStr && utfBytes
               this->gsCols = tempgsCols ;
               this->cWidth[ZERO] = tempcWidth ;
               break ;
            }
         }
      }
   }
}  //* End ColumnCount() *

#if BINARY_EXTENSION != 0
//*************************
//*     binaryFormat      *
//*************************
//******************************************************************************
//* Implements a binary-output extension to the standard swprintf output-format*
//* specifiers.                                                                *
//*                                                                            *
//* Input  : convBuff : receives formatted data                                *
//*          convType : L'b' , L'B' (conversion specifier) or NULLCHAR         *
//*          sepType  : group seperator character                              *
//*          intData  : source data in long-long int format                    *
//*          dataType : data type at *intData: i.e. number of bytes in         *
//*                     source integer                                         *
//*          grpType  : bit grouping: 4 == by-4,  8 == by-8                    *
//*                                                                            *
//*          append   : if 'true, then append the 'b', 'B' (or NULLCHAR)       *
//*                      to end of string                                      *
//*                     if 'false', then prepend the 'b' or 'B' character      *
//*                                                                            *
//* Returns: nothing                                                           *
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//*   Examples:                                                                *
//*    %b   , %B     integer source                                            *
//*    %lb  , %lB    long integer source                                       *
//*    %llb , %llB , %qb , %qB , %Lb , %LB  long long integer (64-bit) source  *
//*    %hb  , %hB    short integer source                                      *
//*    %hhb , %hhB   byte integer source                                       *
//*    %#b  , %#B    prepend a 'b' or 'B' character                            *
//*    %-#b , %-#B   append a 'b' or 'B' character                             *
//*                                                                            *
//*   Output is in groups of four (4) bits (default) or eight (8) bits         *
//*   Group seperator character is '.' (default) or specified character in     *
//*    the range ' ' 20h through '/' 2Fh  OR  ':' 3Ah through '@' 40h  OR      *
//*              '[' 5Bh, ']' 5Dh, '^' 5Eh, '_' 5Fh, '`' 60h,                  *
//*              '{' 7Bh, '|' 7Ch, '}' 7Dh, '~' 7Eh                            *
//*   Examples: %b      1011.0111.1111.0000.0111.0101.1010.0001                *
//*             %#hB   B0111.0101.1010 0001                                    *
//*             %-#hB   0111.0101.1010.0001B                                   *
//*             %-#hhb  0101b                                                  *
//*             % B     1011 0111 1111 0000 0111 0101 1010 0001                *
//*             %-hb    0111-0101-1010-0001                                    *
//*             %#_hb   b0111_0101_1010_0001                                   *
//*             %.4hb   0111.0101.1010.0001                                    *
//*             % .8b   10110111 11110000 01110101 10100001                    *
//*             %-#_.8B 10110111_11110000_01110101_10100001B                   *
//*                                                                            *
//*                                                                            *
//******************************************************************************

void gString::binaryFormat ( wchar_t* convBuff, wchar_t convType, wchar_t sepType, 
                             long long int intData, short dataType, short grpType, 
                             bool append )
{
   short obIndex,          // output-buffer index
         bitCount ;        // number of data bits to process
   switch ( dataType )
   {
      case 1:     obIndex = grpType == 4 ? 10 :  9 ; bitCount =  8 ; break ;
      case 2:     obIndex = grpType == 4 ? 20 : 18 ; bitCount = 16 ; break ;
      case 4:     obIndex = grpType == 4 ? 40 : 36 ; bitCount = 32 ; break ;
      case 8:     obIndex = grpType == 4 ? 80 : 72 ; bitCount = 64 ; break ;
      default:    // programmer error
         obIndex = 1 ;
         bitCount = ZERO ;
         *convBuff = NULLCHAR ;
         break ;
   }
   convBuff[obIndex--] = NULLCHAR ;    // terminate the string
   if ( append != false )
      convBuff[obIndex--] = convType ; // 'b' or 'B' indicator (if specified)
   long long int msk = 0x0000000000000001 ;
   for ( short i = ZERO ; i < bitCount ; i++ )
   {
      if ( i > ZERO && ((i % grpType) == ZERO) )
         convBuff[obIndex--] = sepType ;
      if ( intData & msk ) convBuff[obIndex--] = L'1' ;
      else                 convBuff[obIndex--] = L'0' ;
      msk <<= 1 ;
   }
   if ( append == false )
      convBuff[obIndex] = convType ;

}  //* End binaryFormat() *
#endif   // BINARY_EXTENSION

#if IFORMAT_EXTENSION != 0
//***********************
//*     intFormat32     *
//***********************
//******************************************************************************
//* Private Methods: Called only by 'formatInteger' methods.                   *
//* Convert an integer value to a fixed-width formatted string.                *
//*                                                                            *
//* Input  : wsrc  : buffer to receive formatted data                          *
//*          iVal  : value to be formatted                                     *
//*                  unsigned long (32-bit) OR unsigned long long (64-bit)     *
//*          fWidth: field width (number of display columns)                   *
//*          lJust : (optional, false by default)                              *
//*                  if true, strip leading spaces to left-justify the value   *
//*                  in the field. (string may be less than fWidth)            *
//*          sgnVal: (optional, false by default) [CURRENTLY IGNORED]          *
//*                  if true, the interpret 'iVal' as a signed number          *
//*          pnSign: (optional, false by default) [CURRENTLY IGNORED]          *
//*                  if 'sgnVal' == true AND pnSign == true, the prepend a     *
//*                  '+' or '-' sign to the value. The sign reduces precision  *
//*                  by one digit                                              *
//*                                                                            *
//* Returns: 'true' if successful                                              *
//*          'false' if field overflow (field will be filled with '#' chars)   *
//*                     note that oveflow can occur ONLY:                      *
//*                     if fWidth <= 3 OR if iVal >= 10. terabytes             *
//******************************************************************************

bool gString::intFormat32 ( wchar_t* wsrc, unsigned long iVal, short fWidth, 
                            bool lJust, bool sgnVal, bool pnSign )
{
   bool   overflow = false ;     // return value
   double decVal ;

   if ( fWidth > FI_MAX_FIELDWIDTH )   // limit field width to 32-bit size
      fWidth = FI_MAX_FIELDWIDTH ;
   else if ( fWidth < 1 )              // prevent caller stupidity
      fWidth = 1 ;

   //* For the easy combinations, we let the swprintf() formatting do the work.*
   //* Note that we rely on the LC_NUMERIC locale to be the default 'C' locale.*
   //* If it is not, then some values will be formatted for another locale,    *
   //* but the width should still be correct.                                  *
   if ( (fWidth == FI_MAX_FIELDWIDTH) ||
        (iVal <= 999999999 && fWidth >= 11) ||
        (iVal <=  99999999 && fWidth >= 10) ||
        (iVal <=   9999999 && fWidth >=  9) ||
        (iVal <=    999999 && fWidth >=  7) ||
        (iVal <=     99999 && fWidth >=  6) ||
        (iVal <=      9999 && fWidth >=  5) ||
        (iVal <=      9999 && fWidth ==  4) ||
        (iVal <=       999 && fWidth ==  3) ||
        (iVal <=        99 && fWidth ==  2) ||
        (iVal <=         9 && fWidth ==  1)
      )
   {
      const wchar_t* fTemplate[] = 
      {
         L"%lu",           // (placeholder)
         L"%lu",           // width == 1
         L"%2lu",          // width == 2
         L"%3lu",          // width == 3
         L"%4lu",          // width == 4
         L"%'5lu",         // width == 5
         L"%'6lu",         // width == 6
         L"%'7lu",         // width == 7
         L"%'8lu",         // width == 8
         L"%'9lu",         // width == 9
         L"%'10lu",        // width == 10
         L"%'11lu",        // width == 11
         L"%'12lu",        // width == 12
         L"%'13lu"         // width == 13
      } ;
      swprintf ( wsrc, (FI_MAX_FIELDWIDTH + 1), fTemplate[fWidth], iVal ) ;
   }

   //* For values that will not fit into the field in fully-expanded form      *
   else if ( fWidth == 12 )
   {
      decVal = ((double)iVal) / 1000000000.0 ;
      swprintf ( wsrc, 13, L"%1.9lfG", decVal ) ;
   }
   else if ( fWidth == 11 )
   {
      decVal = ((double)(iVal/10)) / 100000000.0 ;
      swprintf ( wsrc, 12, L"%1.8lfG", decVal ) ;
   }
   else if ( fWidth == 10 )
   {
      const wchar_t* fTemplate[] = 
      {
         L"%3.5lfM",       // val <= 999,999,999
         L"%1.7lfG"        // val >= 1,000,000,000
      } ;
      short fIndex = 1 ;
      if ( iVal <= 999999999 )
      { decVal = ((double)(iVal/10)) / 100000.0 ; fIndex = 0 ; }
      else     // > 999999999
      { decVal = ((double)(iVal/100)) / 10000000.0 ; fIndex = 1 ; }
      swprintf ( wsrc, 11, fTemplate[fIndex], decVal ) ;
   }
   else if ( fWidth == 9 )
   {
      const wchar_t* fTemplate[] = 
      {
         L"%2.5lfM",       // val <= 99,999,999
         L"%3.4lfM",       // val <= 999,999,999
         L"%1.6lfG",       // val >= 1,000,000,000
      } ;
      short fIndex = 2 ;
      if ( iVal <= 99999999 )
      { decVal = ((double)(iVal/10)) / 100000.0 ; fIndex = 0 ; }
      else if ( iVal <= 999999999 )
      { decVal = ((double)(iVal/100)) / 10000.0 ; fIndex = 1 ; }
      else     // > 999999999
      { decVal = ((double)(iVal/1000)) / 1000000.0 ; fIndex = 2 ; }
      swprintf ( wsrc, 10, fTemplate[fIndex], decVal ) ;
   }
   else if ( fWidth == 8 )
   {
      const wchar_t* fTemplate[] = 
      {
         L"%1.5lfM",        // val <= 9,999,999
         L"%2.4lfM",        // val <= 99,999,999
         L"%3.3lfM",        // val <= 999,999,999
         L"%1.5lfG",        // val >= 1,000,000,000
      } ;
      short fIndex = 3 ;
      if ( iVal <= 9999999 )
      { decVal = ((double)(iVal/10)) / 100000.0 ; fIndex = 0 ; }
      else if ( iVal <= 99999999 )
      { decVal = ((double)(iVal/100)) / 10000.0 ; fIndex = 1 ; }
      else if ( iVal <= 999999999 )
      { decVal = ((double)(iVal/1000)) / 1000.0 ; fIndex = 2 ; }
      else  // > 999999999
      { decVal = ((double)(iVal/10000)) / 100000.0 ; fIndex = 3 ; }
      swprintf ( wsrc, 9, fTemplate[fIndex], decVal ) ;
   }
   else if ( fWidth == 7 )
   {
      const wchar_t* fTemplate[] = 
      {
         L"%1.4lfM",       // val <= 9,999,999
         L"%2.3lfM",       // val <= 99,999,999
         L"%3.2lfM",       // val <= 999,999,999
         L"%1.4lfG",       // val >= 1,000,000,000
      } ;
      short fIndex = 3 ;
      if ( iVal <= 9999999 )
      { decVal = ((double)(iVal/100)) / 10000.0 ; fIndex = 0 ; }
      else if ( iVal <= 99999999 )
      { decVal = ((double)(iVal/1000)) / 1000.0 ; fIndex = 1 ; }
      else if ( iVal <= 999999999 )
      { decVal = ((double)(iVal/10000)) / 100.0 ; fIndex = 2 ; }
      else  // > 999999999
      { decVal = ((double)(iVal/100000)) / 10000.0 ; fIndex = 3 ; }
      swprintf ( wsrc, 8, fTemplate[fIndex], decVal ) ;
   }
   else if ( fWidth == 6 )
   {
      const wchar_t* fTemplate[] = 
      {
         L"%3.1lfK",       // val <= 999,999
         L"%1.3lfM",       // val <= 9,999,999
         L"%2.2lfM",       // val <= 99,999,999
         L"%3.1lfM",       // val <= 999,999,999
         L"%1.3lfG",       // val >= 1,000,000,000
      } ;
      short fIndex = 4 ;
      if ( iVal <= 999999 )
      { decVal = ((double)(iVal/100)) / 10.0 ; fIndex = 0 ; }
      else if ( iVal <= 9999999 )
      { decVal = ((double)(iVal/1000)) / 1000.0 ; fIndex = 1 ; }
      else if ( iVal <= 99999999 )
      { decVal = ((double)(iVal/10000)) / 100.0 ; fIndex = 2 ; }
      else if ( iVal <= 999999999 )
      { decVal = ((double)(iVal/100000)) / 10.0 ; fIndex = 3 ; }
      else
      { decVal = ((double)(iVal/1000000)) / 1000.0 ; fIndex = 4 ; }
      swprintf ( wsrc, 7, fTemplate[fIndex], decVal ) ;
   }
   else if ( fWidth == 5 )
   {
      const wchar_t* fTemplate[] = 
      {
         L"%2.1lfK",       // val <= 99,999
         L"% 3.0lfK",      // val <= 999,999
         L"%1.2lfM",       // val <= 9,999,999
         L"%2.1lfM",       // val <= 99,999,999
         L"% 3.0lfM",      // val <= 999,999,999
         L"%1.2lfG"        // val >  999,999,999
      } ;
      short fIndex = 5 ;
      if ( iVal <= 99999 )
      { decVal = ((double)(iVal/100)) / 10.00 ; fIndex = 0 ; }
      else if ( iVal <= 999999 )
      { decVal = ((double)(iVal/1000)) ; fIndex = 1 ; }
      else if ( iVal <= 9999999 )
      { decVal = ((double)(iVal/10000)) / 100.0 ; fIndex = 2 ; }
      else if ( iVal <= 99999999 )
      { decVal = ((double)(iVal/100000)) / 10.0 ; fIndex = 3 ; }
      else if ( iVal <= 999999999 )
      { decVal = ((double)(iVal/1000000)) ; fIndex = 4 ; }
      else
      { decVal = ((double)(iVal/10000000)) / 100.0 ; fIndex = 5 ; }
      swprintf ( wsrc, 6, fTemplate[fIndex], decVal ) ;
   }
   else if ( fWidth == 4 )
   {
      const wchar_t* fTemplate[] = 
      {
         L"% 2.0lfK",      // val <= 99,999
         L"%3.0lfK",       // val <= 999,999
         L"%1.1lfM",       // val <= 9,999,999
         L"% 2.0lfM",      // val <= 99,999,999
         L"%3.0lfM",       // val <= 999,999,999
         L"%1.1lfG"        // val >  999,999,999
      } ;
      short fIndex = 5 ;
      if ( iVal <= 99999 )
      { decVal = ((double)(iVal/1000)) ; fIndex = 0 ; }
      else if ( iVal <= 999999 )
      { decVal = ((double)(iVal/1000)) ; fIndex = 1 ; }
      else if ( iVal <= 9999999 )
      { decVal = ((double)(iVal/100000)) / 10.0 ; fIndex = 2 ; }
      else if ( iVal <= 99999999 )
      { decVal = ((double)(iVal/1000000)) ; fIndex = 3 ; }
      else if ( iVal <= 999999999 )
      { decVal = ((double)(iVal/1000000)) ; fIndex = 4 ; }
      else
      { decVal = ((double)(iVal/1000000)) / 1000.0 ; fIndex = 5 ; }
      swprintf ( wsrc, 5, fTemplate[fIndex], decVal ) ;
   }
   else if ( fWidth == 3 && iVal <= 99999 )
   {
      if ( iVal <= 9999 )
      { decVal = ((double)(iVal/100)) / 10.0 ; }
      else
      { decVal = ((double)(iVal/1000)) ; }
      swprintf ( wsrc, 4, L"%2.0lfK", decVal ) ;
   }
   else     // field overflow
   {
      if ( fWidth == 3 )  wcsncpy ( wsrc, L"###", 4 ) ;
      else if ( fWidth == 2 )  wcsncpy ( wsrc, L"##", 3 ) ;
      //* width == 1 OR caller is a moron *
      else  wcsncpy ( wsrc, L"#", 2 ) ;
      overflow = true ;
   }

   //* If user has specifed stripping the leading spaces *
   if ( lJust != false )
   {
      short s = ZERO, d = ZERO ;
      while ( wsrc[s] == SPACE && wsrc[s] != NULLCHAR )  // point to source
         ++s ;
      if ( s > ZERO && wsrc[s] != NULLCHAR )
      {
         while ( wsrc[s] != NULLCHAR )
            wsrc[d++] = wsrc[s++] ;
         wsrc[d] = NULLCHAR ;
      }
   }

   return overflow ;

}  //* End intFormat32() *

//***********************
//*     intFormat64     *
//***********************
//******************************************************************************
//* Private Methods: Called only by 'formatInteger' methods.                   *
//* This is the 64-bit conversion. See intFormat32() method above.             *
//******************************************************************************

bool gString::intFormat64 ( wchar_t* wsrc, unsigned long long iVal, short fWidth, 
                            bool lJust, bool sgnVal, bool pnSign )
{
   bool  overflow = false ;     // return value
   const unsigned long long int
         max32BIT = (0xFFFFFFFF),
         oneGB = (1000000000),
         tenGB = (oneGB * 10),
         hunGB = (tenGB * 10),
         oneTB = (hunGB * 10),
         tenTB = (oneTB * 10) ;
   unsigned long ulVal ;

   if ( fWidth > FI_MAX_FIELDWIDTH )   // limit field width to 32-bit size
      fWidth = FI_MAX_FIELDWIDTH ;
   else if ( fWidth < 1 )              // prevent caller stupidity
      fWidth = 1 ;

   //* If the value will fit into an unsigned long, *
   //* just call the primary method to handle it.   *
   if ( iVal <= max32BIT )
   {
      ulVal = (unsigned long int)iVal ;
      overflow = this->intFormat32 ( wsrc, ulVal, fWidth, lJust, sgnVal, pnSign ) ;
   }
   else if ( iVal < tenTB )      // if <= our maximum supported value
   {
      double decVal ;
      if ( fWidth == FI_MAX_FIELDWIDTH )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.10lfG",      // 
            L"%2.9lfG",       // 
            L"%3.8lfG",       //
            L"%1.10lfT",      // 
         } ;
         short fIndex = 3 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal)) / 1000000000.000) ; fIndex = 0 ; }
         else if ( iVal < hunGB )
         { decVal = ((double(iVal)) / 1000000000.000) ; fIndex = 1 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / 10)) / 100000000.000) ; fIndex = 2 ; }
         else
         { decVal = ((double(iVal / 100)) / 10000000000.000) ; fIndex = 3 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 12 )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.9lfG",       // <  10 gBytes
            L"%2.8lfG",       // < 100 gBytes
            L"%3.7lfG",       // <   1 tByte
            L"%1.9lfT",       // >=  1 tByte
         } ;
         short fIndex = 3 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal)) / 1000000000.000) ; fIndex = 0 ; }
         else if ( iVal < hunGB )
         { decVal = ((double(iVal / 10)) / 100000000.000) ; fIndex = 1 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / 100)) / 10000000.000) ; fIndex = 2 ; }
         else
         { decVal = ((double(iVal / 1000)) / 1000000000.000) ; fIndex = 3 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 11 )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.8lfG",       // <  10 gBytes
            L"%2.7lfG",       // < 100 gBytes
            L"%3.6lfG",       // <   1 tByte
            L"%1.8lfT",       // >=  1 tByte
         } ;
         short fIndex = 3 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal / 10)) / 100000000.000) ; fIndex = 0 ; }
         else if ( iVal < hunGB )
         { decVal = ((double(iVal / 100)) / 10000000.000) ; fIndex = 1 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / 1000)) / 1000000.000) ; fIndex = 2 ; }
         else
         { decVal = ((double(iVal / 10000)) / 100000000.000) ; fIndex = 3 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 10 )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.7lfG",       // <  10 gBytes
            L"%2.6lfG",       // < 100 gBytes
            L"%3.5lfG",       // <   1 tByte
            L"%1.7lfT",       // >=  1 tByte
         } ;
         short fIndex = 3 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal / 100)) / 10000000.000) ; fIndex = 0 ; }
         else if ( iVal < hunGB )
         { decVal = ((double(iVal / 1000)) / 1000000.000) ; fIndex = 1 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / 10000)) / 100000.000) ; fIndex = 2 ; }
         else
         { decVal = ((double(iVal / 100000)) / 10000000.000) ; fIndex = 3 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 9 )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.6lfG",       // <  10 gBytes
            L"%2.5lfG",       // < 100 gBytes
            L"%3.4lfG",       // <   1 tByte
            L"%1.6lfT",       // >=  1 tByte
         } ;
         short fIndex = 3 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal / 1000)) / 1000000.000) ; fIndex = 0 ; }
         else if ( iVal < hunGB )
         { decVal = ((double(iVal / 10000)) / 100000.000) ; fIndex = 1 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / 100000)) / 10000.000) ; fIndex = 2 ; }
         else
         { decVal = ((double(iVal / 1000000)) / 1000000.000) ; fIndex = 3 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 8 )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.5lfG",       // <  10 gBytes
            L"%2.4lfG",       // < 100 gBytes
            L"%3.3lfG",       // <   1 tByte
            L"%1.5lfT",       // >=  1 tByte
         } ;
         short fIndex = 3 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal / 10000)) / 100000.000) ; fIndex = 0 ; }
         else if ( iVal < hunGB )
         { decVal = ((double(iVal / 100000)) / 10000.000) ; fIndex = 1 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / 1000000)) / 1000.000) ; fIndex = 2 ; }
         else
         { decVal = ((double(iVal / 10000000)) / 100000.000) ; fIndex = 3 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 7 )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.4lfG",       // <  10 gBytes
            L"%2.3lfG",       // < 100 gBytes
            L"%3.2lfG",       // <   1 tByte
            L"%1.4lfT",       // >=  1 tByte
         } ;
         short fIndex = 3 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal / 100000)) / 10000.000) ; fIndex = 0 ; }
         else if ( iVal < hunGB )
         { decVal = ((double(iVal / 1000000)) / 1000.000) ; fIndex = 1 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / 10000000)) / 100.000) ; fIndex = 2 ; }
         else
         { decVal = ((double(iVal / 100000000)) / 10000.000) ; fIndex = 3 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 6 )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.3lfG",       // <  10 gBytes
            L"%2.2lfG",       // < 100 gBytes
            L"%3.1lfG",       // <   1 tByte
            L"%1.3lfT",       // >=  1 tByte
         } ;
         short fIndex = 3 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal / 1000000)) / 1000.000) ; fIndex = 0 ; }
         else if ( iVal < hunGB )
         { decVal = ((double(iVal / 10000000)) / 100.000) ; fIndex = 1 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / 100000000)) / 10.000) ; fIndex = 2 ; }
         else
         { decVal = ((double(iVal / oneGB)) / 1000.000) ; fIndex = 3 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 5 )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.2lfG",       // <  10 gBytes
            L"%2.1lfG",       // < 100 gBytes
            L"%4.0lfG",       // <   1 tByte
            L"%1.2lfT",       // >=  1 tByte
         } ;
         short fIndex = 3 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal / 10000000)) / 100.000) ; fIndex = 0 ; }
         else if ( iVal < hunGB )
         { decVal = ((double(iVal / 100000000)) / 10.000) ; fIndex = 1 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / oneGB))) ; fIndex = 2 ; }
         else
         { decVal = ((double(iVal / tenGB)) / 100.000) ; fIndex = 3 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 4 )
      {
         const wchar_t* fTemplate[] = 
         {
            L"%1.1lfG",       // <  10 gBytes
            L"%3.0lfG",       // <   1 tByte
            L"%1.1lfT",       // >=  1 tByte
         } ;
         short fIndex = 2 ;
         if ( iVal < tenGB )
         { decVal = ((double(iVal / 100000000)) / 10.000) ; fIndex = 0 ; }
         else if ( iVal < oneTB )
         { decVal = ((double(iVal / oneGB))) ; fIndex = 1 ; }
         else
         { decVal = ((double(iVal / hunGB)) / 10.000) ; fIndex = 2 ; }
         swprintf ( wsrc, (fWidth + 1), fTemplate[fIndex], decVal ) ;
      }
      else if ( fWidth == 3 && iVal < hunGB )
      {
         decVal = ((double(iVal / oneGB))) ;
         swprintf ( wsrc, (fWidth + 1), L"%2.0lfG", decVal ) ;
      }
      else
      {
         wcsncpy ( wsrc, L"#############", 14 ) ;
         wsrc[fWidth] = NULLCHAR ;
         overflow = true ;
      }

      //* If user has specifed stripping the leading spaces. *
      //* Note that most values are full field width already.*
      if ( lJust != false && wsrc[0] == SPACE )
      {
         short s = ZERO, d = ZERO ;
         while ( wsrc[s] == SPACE && wsrc[s] != NULLCHAR )  // point to source
            ++s ;
         if ( s > ZERO && wsrc[s] != NULLCHAR )
         {
            while ( wsrc[s] != NULLCHAR )
               wsrc[d++] = wsrc[s++] ;
            wsrc[d] = NULLCHAR ;
         }
      }
   }  // else(val<tenTB)
   else  // values >= 10 terabytes not supported
   {
      wcsncpy ( wsrc, L"#############", 14 ) ;
      wsrc[fWidth] = NULLCHAR ;
      overflow = true ;
   }
   return overflow ;

}  //* End intFormat64() *
#endif   // IFORMAT_EXTENSION

//***************************
//* composeFieldWidthBugFix *
//***************************
//******************************************************************************
//* Compensate for a logical error in the standard library swprintf() function *
//* string conversions.                                                        *
//*                                                                            *
//* Input  : convType : conversion-type code (currently unused)                *
//*          wsrcPtr  : target buffer                                          *
//*          wsrcSpace: free space remaining in buffer                         *
//*          fmt      : output-formatting specification "%s" or "%S"           *
//*                     which includes a field-width specification             *
//*          gsSrc    : source string                                          *
//*                                                                            *
//* Returns: number of characters written to target                            *
//******************************************************************************
//* Due to a bug in the stdlib 'swprintf' function, we have modified the output*
//* formatting for the 'compose' method to compensate for the following        *
//* scenario. The 'swprintf' incorrectly interprets the field-width            *
//* specification as number-of-characters, rather than as number-of-columns as *
//* it should. This is not a problem so long as 1 character == 1 column, with  *
//* Arabic numbers for instance.                                               *
//* However, for conversion specifiers, where the source data contains         *
//* characters that require more than one display column, anomalous formatting *
//* will occur. Specifically, if display columns > character count, then the   *
//* output string will be too wide.                                            *
//*                                                                            *
//* Specifically, for the "%s" and "%S" conversion specifiers that contain a   *
//* a field-width specification, we must manually interpret the field width as *
//* number of display columns. We don't want to do this because it is a        *
//* performance hit AND because some users may actually rely upon the          *
//* unfortunate design choice in the printf family of functions.               *
//*                                                                            *
//* For these reasons, we continue to use the 'swprintf' output directly for   *
//* "%s" and "%S" conversions if:                                              *
//*  a) no field-width specifier is provided (handled by caller), or           *
//*  b) the source data contains only 1-character == 1-column data             *
//*                                                                            *
//* We manually create output of the appropriate column width for the following*
//* set of circumstances.                                                      *
//*  1) if a field-width specification is provided with a string formatting    *
//*     spec, and                                                              *
//*  2) if the provided string contains characters which require more than one *
//*     display column.                                                        *
//* We then (rather redundantly) call 'swprintf' to copy the fully-formated    *
//* string to the output buffer.                                               *
//*                                                                            *
//* NOTE: In future, we may want to use this algorithm for character data      *
//*       as well (%5c  %-2C  etc) or even for numeric data, but at this time  *
//*       it seems wasteful and unnecessary.                                   *
//*                                                                            *
//******************************************************************************

short gString::composeFieldwidthBugFix ( char convType, wchar_t* wsrcPtr, 
                                         short wsrcSpace, 
                                         const wchar_t* fmt, gString& gsSrc )
{
   short swret = ZERO,
         fWidth, absfWidth,
         srcCols = gsSrc.gscols(),
         srcChars = gsSrc.gschars() ;
   swscanf ( fmt, L"%%%hd", &fWidth ) ;     // get the field width
   absfWidth = fWidth >= ZERO ? fWidth : (~fWidth) + 1 ;
   gString gsfmt ;
   gsfmt.compose( L"%%%hdS", &fWidth ) ;

   if ( (srcChars - 1) == (srcCols) )
   {
      swret = swprintf ( wsrcPtr, wsrcSpace, gsfmt.gstr(), gsSrc.gstr() ) ;
   }
   else
   {
      wchar_t wbuf[gsMAXCHARS] ;
      if ( fWidth < ZERO )          // left-justified
      {
         gsSrc.copy( wbuf, srcChars ) ;
         short endSrc = srcChars - 1 + absfWidth - srcCols ;
         for ( short i = (srcChars - 1) ; i < endSrc ; i++ )
            wbuf[i] = SPACE ;
         wbuf[absfWidth] = NULLCHAR ;
      }
      else                          // right-justified
      {
         short endSrc = absfWidth - srcCols ;
         for ( short i = ZERO ; i < endSrc ; i++ )
            wbuf[i] = SPACE ;
         gsSrc.copy( &wbuf[endSrc], srcChars ) ;
      }
      swret = swprintf ( wsrcPtr, wsrcSpace, L"%S", wbuf ) ;
   }
   return swret ;

}  //* End composeFieldwidthBugFix() *

//******************************************************************************
#if ENABLE_GS_DEBUG != 0
//*************************
//*         dbMsg         *
//*************************
//******************************************************************************
//* FOR DEBUGGING ONLY! Application can retrieve most recent debug message.    *
//* Input  : (caller's object, by reference)                                   *
//* Returns: nothing                                                           *
//******************************************************************************

void gString::dbMsg ( gString& gsmsg )
{
   gsmsg = dbmsg ;
}  //* End dbMsg() *
#endif   // ENABLE_GS_DEBUG

//*************************
//*                       *
//*************************
//******************************************************************************
//*                                                                            *
//*                                                                            *
//*                                                                            *
//* Input  :                                                                   *
//*                                                                            *
//* Returns:                                                                   *
//******************************************************************************


